from typing import Optional
from tensorkube.constants import REGION, DEFAULT_NAMESPACE, get_cluster_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.services.nydus import get_nydus_image_url
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name, get_s3_claim_name, get_efs_claim_name
from pkg_resources import resource_filename
import yaml
import click
from kubernetes import config, utils
from tensorkube.services.iam_service import create_or_update_iam_role_with_service_account_cluster_access, attach_role_policy, delete_iam_role
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.k8s_service import patch_service_account, create_service_account

def get_buildkit_command(sanitised_project_name: str, image_tag: str, image_url: Optional[str] =None, upload_to_nfs: bool= False, region: str = REGION, convert_to_nydus_image: bool = False):
    aws_account_number =  get_aws_account_id()
    command = [ "/bin/sh", "-c",
        f"""
        apk add --no-cache curl unzip aws-cli docker
        aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {aws_account_number}.dkr.ecr.{region}.amazonaws.com
        buildctl-daemonless.sh build\
            --frontend dockerfile.v0\
            --local context=/data/build/{sanitised_project_name}\
            --local dockerfile=/data/build/{sanitised_project_name}\
            --output type=image,name={image_url},push=true
        """
    ]
    if convert_to_nydus_image:
        image_repo = image_url.split(":")[0]
        cache_image_url = f"{image_repo}:nydus-buildcache"
        nydus_image_url = get_nydus_image_url(image_url=image_url)
        command = [ "/bin/sh", "-c",
            f"""
            apk add --no-cache curl unzip aws-cli docker
            aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {aws_account_number}.dkr.ecr.{region}.amazonaws.com
            time buildctl-daemonless.sh build\
                --frontend dockerfile.v0\
                --local context=/data/build/{sanitised_project_name}\
                --local dockerfile=/data/build/{sanitised_project_name}\
                --output type=image,name={nydus_image_url},push=true,compression=nydus,force-compression=true,oci-mediatypes=true
            """
        ]
        return command

    if upload_to_nfs:
        command =  ["/bin/sh", "-c", f"""
        mkdir -p /data/cache
        mkdir -p /test/tars/{sanitised_project_name} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs /test/tars/{sanitised_project_name} /test/images/{sanitised_project_name}/rootfs
        time buildctl-daemonless.sh build\
            --frontend dockerfile.v0\
            --local context=/data/build/{sanitised_project_name}\
            --local dockerfile=/data/build/{sanitised_project_name}\
            --output type=tar,dest=/test/tars/{sanitised_project_name}/{sanitised_project_name}-{image_tag}.tar
        if [ $? -ne 0 ]; then
            echo "buildctl-daemonless.sh command failed"
            exit 1
        fi
        echo "Extracting rootfs from tar file"
        time tar -xf /test/tars/{sanitised_project_name}/{sanitised_project_name}-{image_tag}.tar -C /test/images/{sanitised_project_name}/rootfs --checkpoint=1000 --checkpoint-action=echo="Extracted #%u: %T"
        if [ $? -ne 0 ]; then
            echo "extracting rootfs from tar file failed"
            exit 1
        fi
        echo "Extracted rootfs from tar file"
        echo "Uploading rootfs to NFS"
        cd /test/images/{sanitised_project_name}/rootfs
        find . -type d | split -l 5000 - batch_dir_
        find . -type f | split -l 5000 - batch_file_
        find . -type l | split -l 5000 - batch_link_

        total_batches=$(ls batch_* | wc -l)
        processed_batches=0
        total_time=0

        process_batches() {{
            local batch_type=$1
            for batch_file in batch_${{batch_type}}_*; do
                start_time=$(date +%s)
                
                if [ "$batch_type" == "dir" ]; then
                    cat $batch_file | parallel -j 15 mkdir -p /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/{{}}
                elif [ "$batch_type" == "file" ]; then
                    cat $batch_file | parallel -j 15 cp --parents {{}} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/
                elif [ "$batch_type" == "link" ]; then
                    cat $batch_file | parallel -j 15 cp --parents -P {{}} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/
                fi
                
                end_time=$(date +%s)
                batch_time=$((end_time - start_time))
                total_time=$((total_time + batch_time))
                processed_batches=$((processed_batches + 1))
                remaining_batches=$((total_batches - processed_batches))
                avg_time_per_batch=$((total_time / processed_batches))
                estimated_remaining_time=$((avg_time_per_batch * remaining_batches))
                
                echo "Processed batch $processed_batches/$total_batches. Remaining: $remaining_batches. Time for this batch: $batch_time seconds. Estimated remaining time: $estimated_remaining_time seconds."
                rm $batch_file
            done
        }}

        process_batches "dir"
        process_batches "file"
        process_batches "link"
        echo "Uploaded rootfs to NFS" """]
    return command


def apply_k8s_buildkit_config(sanitised_project_name: str, image_tag: str, region: str = REGION,
                              env_name: Optional[str] = None, context_name: Optional[str] = None, image_url: Optional[str] = None, upload_to_nfs: bool = False,convert_to_nydus_image: bool = False):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    buildkit_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/buildkit.yaml')
    with open(buildkit_config_file_path) as f:
        buildkit_config = yaml.safe_load(f)

    buildkit_config['metadata']['name'] = 'buildkit-{}'.format(sanitised_project_name)
    buildkit_config['spec']['template']['spec']['containers'][0]['env'][0]['value'] = region
    buildkit_config['spec']['ttlSecondsAfterFinished'] = 100
    

    # Include the namespace in the buildkit configuration
    # Replace 'default' with your default namespace if needed
    namespace_to_use = env_name if env_name else DEFAULT_NAMESPACE
    buildkit_config['metadata']['namespace'] = namespace_to_use
    
    # set service account name so that pod can access ECR   
    buildkit_config['spec']['template']['spec']['serviceAccountName'] = get_buildkit_service_account_name(namespace=namespace_to_use)

    # Modify volume claim names based on env_name
    for volume in buildkit_config['spec']['template']['spec']['volumes']:
        if volume['name'] == 'persistent-storage':
            volume['persistentVolumeClaim']['claimName'] = get_s3_claim_name(env_name=env_name)
        elif volume['name'] == 'efs-pvc':
            volume['persistentVolumeClaim']['claimName'] = get_efs_claim_name(env_name=env_name)

    buildkit_config['spec']['template']['spec']['containers'][0]['command'] = get_buildkit_command(sanitised_project_name, image_tag, image_url, upload_to_nfs, region, convert_to_nydus_image)

    utils.create_from_dict(k8s_api_client, buildkit_config)
    click.echo('Deployed a Buildkit image')


def configure_buildkit_irsa(env_name: Optional[str] = None):
    namespace = env_name if env_name else DEFAULT_NAMESPACE
    service_account_name = get_buildkit_service_account_name(namespace=namespace)
    create_service_account(service_account_name, namespace)
    role_name = get_buildkit_ecr_iam_role_name()
    role = create_or_update_iam_role_with_service_account_cluster_access(get_aws_account_id(),get_cluster_oidc_issuer_url(get_cluster_name()), role_name, service_account_name, namespace)
    role_arn = role["Role"]["Arn"]
    attach_role_policy("aws", "AmazonEC2ContainerRegistryPowerUser", role_name, REGION)
    patch_service_account(role_arn, service_account_name, namespace)

def get_buildkit_service_account_name(namespace: str):
    return f"{get_cluster_name()}-{namespace}-buildkit-sa"

def get_buildkit_ecr_iam_role_name():
    return f"{get_cluster_name()}-buildkit-ecr-role"


def delete_buildkit_irsa():
    role_name = get_buildkit_ecr_iam_role_name()
    delete_iam_role(role_name)
    click.echo(f"Deleted Buildkit IAM role '{role_name}'")


import enum
import pprint
import threading
import time
from typing import Tuple, Optional

import botocore.exceptions
import click
from pkg_resources import resource_filename

from tensorkube.services.aws_service import get_cloudformation_client, get_ec2_client, get_iam_client
from tensorkube.constants import get_cluster_name, TENSORFUSE_STRING, CREATED_BY_TAG, CLUSTER_NAME_TAG


class CfnStackStatus(enum.Enum):
    CREATE_IN_PROGRESS = 'CREATE_IN_PROGRESS'
    CREATE_COMPLETE = 'CREATE_COMPLETE'
    CREATE_FAILED = 'CREATE_FAILED'
    ROLLBACK_IN_PROGRESS = 'ROLLBACK_IN_PROGRESS'
    ROLLBACK_COMPLETE = 'ROLLBACK_COMPLETE'
    ROLLBACK_FAILED = 'ROLLBACK_FAILED'
    UPDATE_IN_PROGRESS = 'UPDATE_IN_PROGRESS'
    UPDATE_COMPLETE = 'UPDATE_COMPLETE'
    UPDATE_FAILED = 'UPDATE_FAILED'
    DELETE_IN_PROGRESS = 'DELETE_IN_PROGRESS'
    DELETE_COMPLETE = 'DELETE_COMPLETE'
    DELETE_FAILED = 'DELETE_FAILED'
    UPDATE_ROLLBACK_IN_PROGRESS = 'UPDATE_ROLLBACK_IN_PROGRESS'
    UPDATE_ROLLBACK_COMPLETE = 'UPDATE_ROLLBACK_COMPLETE'
    UPDATE_ROLLBACK_FAILED = 'UPDATE_ROLLBACK_FAILED'
    REVIEW_IN_PROGRESS = 'REVIEW_IN_PROGRESS'


def get_stack_status_from_stack_name(stack_name: str) -> CfnStackStatus:
    """
    Get the status of a CloudFormation stack.

    Args:
        stack_name (str): Name of the CloudFormation stack

    Returns:
        CfnStackStatus: Enum representing the stack status

    Raises:
        ClientError: If there's an AWS API error other than stack not existing
    """
    cloudformation_client = get_cloudformation_client()
    try:
        response = cloudformation_client.describe_stacks(StackName=stack_name)
        status = response['Stacks'][0]['StackStatus']
        return CfnStackStatus(status)
    except Exception as e:
        if 'does not exist' in str(e):
            return None
        raise


def create_cloudformation_stack()-> Tuple[bool, bool]:
    cloudformation_client = get_cloudformation_client()

    # stack name
    cluster_name = get_cluster_name()
    parameters = [{"ParameterKey": "ClusterName", 'ParameterValue': cluster_name}]

    file_name = resource_filename('tensorkube', 'configurations/karpenter_cloudformation.yaml')

    try:
        stack = cloudformation_client.describe_stacks(StackName=cluster_name)
        if stack['Stacks'][0]['StackStatus'] == 'CREATE_COMPLETE':
            click.echo(f"Stack {cluster_name} already exists. Skipping stack creation")
            creation_queued = False
            in_created_state = True
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'CREATE_IN_PROGRESS':
            click.echo(f"Stack {cluster_name} already exists and is in creation state. Please wait for "
                       f"creation to complete.")
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'ROLLBACK_COMPLETE':
            click.echo(f"Stack {cluster_name} already exists but is in rollback state. Please delete the stack "
                       f"and recreate it.")
            creation_queued = False
            in_created_state = False
            return creation_queued, in_created_state
        elif stack['Stacks'][0]['StackStatus'] == 'DELETE_COMPLETE':
            # use the configurations/karpenter_cloudformation.yaml in the library
            with open(file_name) as file:
                template = file.read()
            response = cloudformation_client.create_stack(StackName=cluster_name, TemplateBody=template, Parameters=parameters,
                                                          Capabilities=["CAPABILITY_NAMED_IAM"])
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        else:
            click.echo(f"Stack {cluster_name} already exists but not in created state. Either wait for "
                       f"creation to complete or delete stack to recreate it.")
            creation_queued = False
            in_created_state = False
            return creation_queued, in_created_state
    except botocore.exceptions.ClientError as e:
        if 'does not exist' in str(e):
            # use the configurations/karpenter_cloudformation.yaml in the library
            with open(file_name) as file:
                template = file.read()
            response = cloudformation_client.create_stack(StackName=cluster_name, TemplateBody=template, Parameters=parameters,
                                                          Capabilities=["CAPABILITY_NAMED_IAM"])
            creation_queued = True
            in_created_state = False
            return creation_queued, in_created_state
        else:
            click.echo(e)
            raise Exception('Unable to create cloudformation cluster')


def delete_role_and_attached_policies(iam, role_name):
    # List all attached policies
    attached_policies = iam.list_attached_role_policies(RoleName=role_name)['AttachedPolicies']

    # Detach each policy
    for policy in attached_policies:
        iam.detach_role_policy(RoleName=role_name, PolicyArn=policy['PolicyArn'])
        if get_cluster_name() in policy['PolicyName']:
            # Delete the policy
            iam.delete_policy(PolicyArn=policy['PolicyArn'])

    # Delete the role
    iam.delete_role(RoleName=role_name)


def delete_cloudformation_stack(stack_name):
    cloudformation_client = get_cloudformation_client()

    try:
        cloudformation_client.describe_stacks(StackName=stack_name)
    except botocore.exceptions.ClientError as e:
        if 'does not exist' in str(e):
            print(f'Stack {stack_name} does not exist.')
            return
        else:
            raise
    iam = get_iam_client()

    role_name = f'KarpenterNodeRole-{get_cluster_name()}'

    # List all instance profiles
    instance_profiles = iam.list_instance_profiles()['InstanceProfiles']

    # For each instance profile
    for profile in instance_profiles:
        # Check if the role is associated with the instance profile
        for role in profile['Roles']:
            if role['RoleName'] == role_name:
                # Remove the role from the instance profile
                iam.remove_role_from_instance_profile(InstanceProfileName=profile['InstanceProfileName'],
                    RoleName=role_name)

                # Delete the instance profile
                iam.delete_instance_profile(InstanceProfileName=profile['InstanceProfileName'])

    # Delete the role
    delete_role_and_attached_policies(iam, role_name)

    response = cloudformation_client.delete_stack(StackName=stack_name)
    # Create a waiter to wait for the stack to be deleted
    waiter = cloudformation_client.get_waiter('stack_delete_complete')

    # Start streaming the stack events in a separate thread
    stream_stack_events(stack_name)

    # Wait for the stack to be deleted
    waiter.wait(StackName=stack_name, WaiterConfig={'MaxAttempts': 30})
    return response


def delete_launch_templates():
    ec2_client = get_ec2_client()
    cluster_name = get_cluster_name()
    # Describe the launch templates
    response = ec2_client.describe_launch_templates(
        Filters=[{'Name': 'tag:karpenter.k8s.aws/cluster', 'Values': [cluster_name]}])
    launch_template_names = [lt['LaunchTemplateName'] for lt in response['LaunchTemplates']]

    # Delete each launch template
    for name in launch_template_names:
        ec2_client.delete_launch_template(LaunchTemplateName=name)


def stream_stack_events(stack_name):
    seen_events = set()
    cf_client = get_cloudformation_client()
    while True:
        try:
            events = cf_client.describe_stack_events(StackName=stack_name)['StackEvents']
            for event in reversed(events):
                event_id = event['EventId']
                if event_id not in seen_events:
                    seen_events.add(event_id)
                    click.echo(
                        f"{event['Timestamp']} {event['ResourceStatus']} {event['ResourceType']} {event['LogicalResourceId']} {event.get('ResourceStatusReason', '')}")
            # Check if stack creation is complete
            stack_status = cf_client.describe_stacks(StackName=stack_name)['Stacks'][0]['StackStatus']
            if stack_status.endswith('_COMPLETE') or stack_status.endswith('_FAILED'):
                break
            time.sleep(5)
        except Exception as e:
            if stack_name in str(e) and 'does not exist' in str(e):
                break


def cloudformation():
    """Create a cloudformation stack."""
    click.echo("Creating cloudformation stack...")
    # create_cloudformation_stack()
    creation_queued, in_created_state = create_cloudformation_stack()
    if creation_queued:
        stream_stack_events(get_cluster_name())
    click.echo("Cloudformation stack created.")


def deploy_generic_cloudformation_stack(stack_name: str, template_body: str, parameters: list,
                                capabilities=None,
                                should_wait: bool = True,
                                enable_termination_protection: bool = False) -> bool:
    """
    Base function to deploy any CloudFormation stack using boto3 waiters
    This method also checks if the stack already exists.
    It only creates a stack if it doesnt exist
    """
    if capabilities is None:
        capabilities = ['CAPABILITY_IAM']
    cloudformation_client = get_cloudformation_client()

    tags = [
        {'Key': CREATED_BY_TAG, 'Value': TENSORFUSE_STRING},
        {'Key': CLUSTER_NAME_TAG, 'Value': get_cluster_name()}
    ]

    try:
        # Check if stack exists
        try:
            stack = cloudformation_client.describe_stacks(StackName=stack_name)
            current_stack = stack['Stacks'][0]
            current_status = current_stack['StackStatus']
            click.echo(f'Stack status: {current_status}')

            if current_status == CfnStackStatus.CREATE_COMPLETE.value:
                return True
            elif current_status in [CfnStackStatus.CREATE_IN_PROGRESS.value,
                                    CfnStackStatus.UPDATE_IN_PROGRESS.value]:
                if should_wait:
                    waiter = cloudformation_client.get_waiter('stack_create_complete')
                    waiter.wait(StackName=stack_name)
                return True
            elif current_status in [CfnStackStatus.ROLLBACK_COMPLETE.value, CfnStackStatus.DELETE_COMPLETE.value]:
                click.echo(click.style(
                    f"Error: Stack {stack_name} has status {current_status}. Please check the CloudFormation console for details.",
                    fg='red'))
                return False
            else:
                click.echo(click.style(
                    f"Unhandled stack status {current_status}. Please take necessary actions.",
                    fg='yellow'))
                return False

        except cloudformation_client.exceptions.ClientError as e:
            if 'does not exist' in str(e):
                # Create new stack
                response = cloudformation_client.create_stack(
                    StackName=stack_name,
                    TemplateBody=template_body,
                    Parameters=parameters,
                    Capabilities=capabilities,
                    EnableTerminationProtection=enable_termination_protection,
                    Tags=tags
                )

                # Wait for stack creation
                if should_wait:
                    waiter = cloudformation_client.get_waiter('stack_create_complete')
                    waiter.wait(StackName=stack_name)
                time.sleep(3)
                # Get final stack details
                final_stack = cloudformation_client.describe_stacks(
                    StackName=stack_name
                )['Stacks'][0]

                return True
            raise
    except Exception as e:
        return False

def get_validation_records_from_events(stack_name: str) -> Optional[dict]:
    """Get DNS validation records from stack events during creation"""
    cloudformation_client = get_cloudformation_client()
    try:
        events = cloudformation_client.describe_stack_events(StackName=stack_name)['StackEvents']
        # check if the stack is in CREATE_IN_PROGRESS state
        stack_status = cloudformation_client.describe_stacks(StackName=stack_name)['Stacks'][0]['StackStatus']
        if stack_status == CfnStackStatus.CREATE_COMPLETE.value:
            click.echo(
                click.style(
                    "Certificate already validated. No need to create a validation record. Only redirection record needed.","green"
                )
            )
            return {}
        for event in events:
            if (event['ResourceType'] == 'AWS::CertificateManager::Certificate' and
                    event['ResourceStatus'] == 'CREATE_IN_PROGRESS' and
                    'Content of DNS Record is:' in event.get('ResourceStatusReason', '')):

                # Extract DNS record details from ResourceStatusReason
                record_info = event['ResourceStatusReason'].split('Content of DNS Record is: ')[1]
                # Remove curly braces and split by comma
                record_parts = record_info.strip('{}').split(',')

                validation_record = {}
                for part in record_parts:
                    key, value = part.split(':', 1)
                    validation_record[key.strip()] = value.strip(' .')

                return {
                    'validation_record': {
                        'name': validation_record['Name'],
                        'value': validation_record['Value'],
                        'type': validation_record['Type']
                    }
                }
    except Exception as e:
        click.echo(click.style(f"Error getting validation records: {str(e)}", fg='red'))
        return None
    return None


def get_validation_records_with_wait(stack_name: str, timeout: int = 300, poll_interval: int = 10) -> Optional[dict]:
    """
    Wait until DNS validation records are available in CloudFormation stack events.

    :param stack_name: Name of the CloudFormation stack
    :param timeout: Maximum duration to wait for records (in seconds)
    :param poll_interval: Time to wait between checks (in seconds)
    :return: Validation records if available, None otherwise
    """
    start_time = time.time()

    while (time.time() - start_time) < timeout:
        records = get_validation_records_from_events(stack_name)
        if records is not None:
            return records
        time.sleep(poll_interval)

    click.echo(click.style("Timeout reached while waiting for DNS validation records.", fg='red'))
    return None



from typing import Optional
import json
import click

from pkg_resources import resource_filename
from tensorkube.constants import DEFAULT_NAMESPACE, REGION
from tensorkube.services.iam_service import create_iam_role_with_service_account_cluster_access, create_iam_policy, \
    attach_role_policy, delete_iam_role, delete_policy, detach_role_policy
from tensorkube.services.eks_service import get_cluster_name, get_cluster_oidc_issuer_url
from tensorkube.services.k8s_service import create_service_account_with_role_arn
from tensorkube.services.aws_service import get_aws_account_id

def create_s3_access_to_pods(env_name: Optional[str] = None, region=REGION):
    cluster_name = get_cluster_name()
    oidc_issuer_url = get_cluster_oidc_issuer_url(cluster_name)
    namespace = env_name if env_name else DEFAULT_NAMESPACE
    role_name = get_s3_access_role_name(env_name=env_name)
    service_account_name = get_s3_service_account_name(env_name=env_name)

    create_iam_role_with_service_account_cluster_access(get_aws_account_id(), oidc_issuer_url, role_name, service_account_name,
                                                        namespace)
    policy_name = get_s3_access_policy_name()
    create_s3_access_policy(policy_name, region=region)
    attach_role_policy(get_aws_account_id(), policy_name=policy_name, role_name=role_name)
    role_arn = f"arn:aws:iam::{get_aws_account_id()}:role/{role_name}"
    create_service_account_with_role_arn(name=service_account_name, namespace=namespace, role_arn=role_arn)

def create_s3_access_policy(policy_name: str, region=REGION):
    policy_file_path = resource_filename('tensorkube', 'configurations/aws_configs/s3_access_policy.json')
    with open(policy_file_path, 'r') as f:
        policy = json.load(f)
    for statement in policy['Statement']:
        statement['Resource'] = [r.replace('BUCKET', f"{get_cluster_name()}-*/*") for r in statement['Resource']]
    return create_iam_policy(policy_name, policy, region)

def get_s3_access_role_name(env_name: Optional[str] = None):
    namespace = env_name if env_name else DEFAULT_NAMESPACE
    return f"{get_cluster_name()}-{namespace}-s3-bucket-access"

def get_s3_service_account_name(env_name: Optional[str] = None):
    namespace = env_name if env_name else DEFAULT_NAMESPACE
    return f"{get_cluster_name()}-{namespace}-pod-s3-bucket-access-sa"

def get_s3_access_policy_name(env_name: Optional[str] = None):
    namespace = env_name if env_name else DEFAULT_NAMESPACE
    return f"{get_cluster_name()}-{namespace}-s3-bucket-access-policy"


def delete_s3_access_to_pods(env_name: Optional[str] = None, region=REGION):
    role_name = get_s3_access_role_name(env_name=env_name)
    policy_name = get_s3_access_policy_name(env_name=env_name)
    try:
        detach_role_policy(get_aws_account_id(), policy_name=policy_name, role_name=role_name)
    except Exception as e:
        click.echo(f"An error occurred while detaching policy from role: {e}")
    try:
        delete_iam_role(role_name)
    except Exception as e:
        click.echo(f"An error occurred while deleting role: {e}")
    try:
        delete_policy(policy_name)
    except Exception as e:
        click.echo(f"An error occurred while deleting policy: {e}")


import socket

import click
import metapod_python
import asyncio




def list_instances_in_cloud(cloud_name: str = None):
    async def run_async():
        return await metapod_python.list_instances_in_cloud(cloud_name)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        # If we're already in an event loop, use create_task
        return asyncio.create_task(run_async())
    else:
        # If no event loop is running, use run_until_complete
        return loop.run_until_complete(run_async())

def is_port_available(port: int) -> bool:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind(('127.0.0.1', port))
        return True
    except OSError:
        return False
    finally:
        s.close()


def start_devcontainer(cloud_name:str, gpu_type:str, port:int):
    if not is_port_available(port):
        # Port is not available. print in red and return
        click.echo(click.style(f"Port {port} is not available. Please choose a different port.", fg='red'))
        return
    async def run_async():
        return await metapod_python.start_devcontainer(cloud_name, port, gpu_type)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        # If we're already in an event loop, use create_task
        return asyncio.create_task(run_async())
    else:
        # If no event loop is running, use run_until_complete
        return loop.run_until_complete(run_async())

def pause_devcontainer(cloud_name:str):
    async def run_async():
        return await metapod_python.pause_devcontainer(cloud_name)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        # If we're already in an event loop, use create_task
        return asyncio.create_task(run_async())
    else:
        # If no event loop is running, use run_until_complete
        return loop.run_until_complete(run_async())

def purge_devcontainer(cloud_name:str):
    async def run_async():
        return await metapod_python.purge_devcontainer(cloud_name)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        # If we're already in an event loop, use create_task
        return asyncio.create_task(run_async())
    else:
        # If no event loop is running, use run_until_complete
        return loop.run_until_complete(run_async())

def reset_cloud(cloud_name:str):
    async def run_async():
        return await metapod_python.reset_cloud(cloud_name)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        # If we're already in an event loop, use create_task
        return asyncio.create_task(run_async())
    else:
        # If no event loop is running, use run_until_complete
        return loop.run_until_complete(run_async())


import inspect
from pathlib import Path
from typing import List

import click

from tensorkube.services.cli import tensorkube

# this file handles automatic doc creation only for the reference section. Try to keep this as simple as possible
# also since the changes might propogate always keep an eye on the doc PR that this thing generates.

class ClickDocGenerator:
    # Blacklist for groups and commands that should be ignored
    BLACKLISTED_GROUPS = {'completion',  # example: ignore completion group
        'internal',  # example: ignore internal group
        'cluster'}

    BLACKLISTED_COMMANDS = {'test',  # example: ignore version command
        'teardown'  # example: ignore help command
    }

    def __init__(self, cli: click.Group):
        self.cli = cli

    def _format_option(self, opt: click.Option) -> str:
        """Format a single option with compact styling."""
        # Get option type
        opt_type = (f"Choice({opt.type.choices})" if isinstance(opt.type,
                                                                click.Choice) else opt.type.name.upper() if opt.type else "STRING")

        # Get default value
        default = str(opt.default) if opt.default is not None else None

        # Get the option name as it appears in CLI (with dashes)
        opt_name = opt.opts[-1]  # Use the last option name (usually the long form)

        # Split help text into main line and additional lines
        help_lines = opt.help.split('\n') if opt.help else ["No description available."]
        main_help = help_lines[0].strip()
        additional_help = [line.strip() for line in help_lines[1:] if line.strip()]

        # Format main option line
        if default:
            option_line = f"* `{opt_name} {opt_type}` [default: {default}]: {main_help}"
        else:
            option_line = f"* `{opt_name} {opt_type}`: {main_help}"

        # Add additional help lines with indentation if they exist
        if additional_help:
            option_line += "\n" + "\n".join(f"    {line}" for line in additional_help)

        return option_line

    def _format_argument(self, arg: click.Argument) -> str:
        """Format a single argument."""
        arg_type = (f"Choice({arg.type.choices})" if isinstance(arg.type,
                                                                click.Choice) else arg.type.name.upper() if arg.type else "STRING")

        # Format argument name in uppercase with angle brackets
        arg_name = f"<{arg.name.upper()}>"

        help_text = arg.help if hasattr(arg, 'help') else "No description available."
        return f"* `{arg_name}` ({arg_type}): {help_text}"

    def _should_include_command(self, cmd_name: str, cmd: click.Command) -> bool:
        """Check if command should be included in documentation."""
        if isinstance(cmd, click.Group):
            return cmd_name not in self.BLACKLISTED_GROUPS
        return cmd_name not in self.BLACKLISTED_COMMANDS

    def _format_command_help(self, cmd: click.Command, path: str = "") -> str:
        """Format help text for a command."""
        doc = cmd.help or inspect.getdoc(cmd)

        # Split docstring into main description and rest
        sections = []

        # Add the heading section for the subcommand
        sections.extend([f"## `{path}`"])

        if doc:
            # Add the main description
            doc_parts = doc.split('Examples:', 1)
            description = doc_parts[0].strip()
            sections.extend([f"{description}\n"])

            # Add Examples section if present
            if len(doc_parts) > 1:
                sections.extend(["**Examples:**\n", doc_parts[1].strip(), "\n"])
        else:
            sections.extend(["No description available.\n"])

        # Add Usage section
        sections.extend(["**Usage:**\n"])

        # Build the usage string
        usage_parts = [path]

        # Add arguments to usage
        arguments = [p for p in cmd.params if isinstance(p, click.Argument)]
        for arg in arguments:
            usage_parts.append(f"<{arg.name.upper()}>")

        # Add options indication to usage if there are options
        options = [p for p in cmd.params if isinstance(p, click.Option)]
        if options:
            usage_parts.append("[OPTIONS]")

        # Add COMMAND [ARGS]... for groups
        if isinstance(cmd, click.Group):
            usage_parts.append("COMMAND [ARGS]...")
        # join all parts
        # Join all parts with spaces
        usage_string = " ".join(usage_parts)

        # Add the usage block with bash identifier
        sections.extend([
            "```bash",
            usage_string,
            "```\n"
        ])


        # Add Arguments section if present
        arguments = [p for p in cmd.params if isinstance(p, click.Argument)]
        if arguments:
            sections.append("**Arguments:**\n")
            sections.extend([self._format_argument(arg) + "\n" for arg in arguments])

        # Add Options section
        options = [p for p in cmd.params if isinstance(p, click.Option)]
        if options or isinstance(cmd, click.Command):
            sections.append("**Options:**\n")
            sections.extend([self._format_option(opt) + "\n" for opt in options])

            # Add help option
            sections.append("* `--help BOOL` [default: false]: Show this message and exit.\n")

        # Add Commands section for groups, filtering out blacklisted commands
        if isinstance(cmd, click.Group):
            sections.append("**Commands:**\n")
            for name, subcmd in cmd.commands.items():
                if self._should_include_command(name, subcmd):
                    help_text = subcmd.help or inspect.getdoc(subcmd)
                    short_help = help_text.split('\n')[0] if help_text else "No description."
                    sections.append(f"* `{name}`: {short_help}\n")

        return "\n".join(sections)

    def generate_docs(self, output_dir: str):
        """Generate documentation for the entire CLI."""
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)

        def process_group(group: click.Group, parents: List[str] = None):
            if parents is None:
                parents = []

            current_path = " ".join([*parents, group.name])

            # Handle base group (tensorkube) commands
            if len(parents) == 0:
                # Generate separate files for direct commands, filtering blacklisted ones
                for cmd_name, cmd in group.commands.items():
                    if not isinstance(cmd, click.Group) and self._should_include_command(cmd_name, cmd):
                        cmd_path = f"{current_path} {cmd_name}"
                        content = ["---", f"title: '{cmd_path}'", f"sidebarTitle: \"{cmd_path}\"", "---\n",
                            self._format_command_help(cmd, cmd_path)]
                        filename = cmd_path.replace(" ", "_").replace("-", "_")
                        (output_path / f"{filename}.mdx").write_text("\n".join(content))

                # Generate base group file with only group info
                content = ["---", f"title: '{current_path}'", f"sidebarTitle: \"{current_path}\"", "---\n",
                    self._format_command_help(group, current_path)]
                filename = current_path.replace(" ", "_").replace("-", "_")
                (output_path / f"{filename}.mdx").write_text("\n".join(content))

            # Handle subgroups (tensorkube <sub-group>)
            elif len(parents) == 1:
                # Skip blacklisted groups
                if not self._should_include_command(group.name, group):
                    return
                content = ["---", f"title: '{current_path}'", f"sidebarTitle: \"{current_path}\"", "---\n",
                    self._format_command_help(group, current_path)]

                # Add subcommand documentation for this group, filtering blacklisted ones
                for cmd_name, cmd in group.commands.items():
                    if self._should_include_command(cmd_name, cmd):
                        cmd_path = f"{current_path} {cmd_name}"
                        content.append("\n" + self._format_command_help(cmd, cmd_path))

                filename = current_path.replace(" ", "_").replace("-", "_")
                (output_path / f"{filename}.mdx").write_text("\n".join(content))

            # Process subgroups, filtering blacklisted ones
            for cmd_name, cmd in group.commands.items():
                if isinstance(cmd, click.Group) and self._should_include_command(cmd_name, cmd):
                    process_group(cmd, [*parents, group.name])

        process_group(self.cli)


if __name__ == '__main__':
    generator = ClickDocGenerator(tensorkube)
    generator.generate_docs("docs/commands")




from tensorkube.services.istio import install_istio_on_cluster
from tensorkube.services.nodepool import get_gpu_nodepool
from tensorkube.services.nydus import install_nydus_snapshotter_helm
import click

def apply():
    try:
        install_istio_on_cluster()
        click.echo("Successfully increased connection idle timeout")
        install_nydus_snapshotter_helm(get_gpu_nodepool())
        click.echo("Successfully installed nydus snapshotter helm")

    except Exception as e:
        raise e



import json
import os
import platform
import subprocess

import click
import semver
from kubernetes import config, client
from pkg_resources import resource_filename

from tensorkube.configurations.configuration_urls import DOMAIN_SERVER_URL, KNATIVE_ISTIO_CONTROLLER_URL
from tensorkube.constants import CliColors, LOCKED_ISTIO_VERSION
from tensorkube.services.eks_service import get_pods_using_namespace, apply_yaml_from_url, delete_resources_from_url
from tensorkube.services.error import CLIVersionError
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name


def check_and_install_istioctl():
    """Check if istioctl is installed and install it if it's not."""
    try:
        output = subprocess.run(["istioctl", "version", "-o", "json"], check=True, stdout=subprocess.PIPE)
        version_dict = json.loads(output.stdout.decode("utf-8"))
        version = semver.VersionInfo.parse(version_dict["clientVersion"]["version"])
        locked_istioctl_version = semver.VersionInfo.parse(LOCKED_ISTIO_VERSION)
        print("istioctl is already installed.")
        if (version.major, version.minor) < (locked_istioctl_version.major, locked_istioctl_version.minor):
            text = f"istioctl version is {version}. Please upgrade istioctl to version above {LOCKED_ISTIO_VERSION}"
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("istioctl not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        if platform.system() == "Darwin" or platform.system() == "Linux":
            try:
                install_command = f"curl -sL https://istio.io/downloadIstioctl | ISTIO_VERSION={LOCKED_ISTIO_VERSION} sh -"
                # Download and install istioctl
                subprocess.run(install_command, shell=True, check=True)
                # the above script installs istioctl in ~/.istioctl/bin/istioctl
                istioctl_path = os.path.expanduser("~/.istioctl/bin/istioctl")
                subprocess.run(f"sudo mv {istioctl_path} /usr/local/bin/", shell=True, check=True)
                print("istioctl moved to /usr/local/bin successfully.")
            except subprocess.CalledProcessError as e:
                print("Unable to install istioctl using curl. Please install istioctl manually.")
                raise e
        else:
            print("Unsupported operating system. Please install istioctl manually.")
            raise Exception('Unsupported operating system.')

        # Verify istioctl installation
        try:
            subprocess.run(["istioctl", "version"], check=True)
            print("istioctl installed successfully.")
        except subprocess.CalledProcessError as e:
            print("istioctl installation failed. Please install istioctl manually.")
            raise e


def install_istio_on_cluster():
    """Install Istio with the default profile."""
    try:
        values_file = resource_filename('tensorkube','configurations/istio/istio.yaml')
        subprocess.run(["istioctl", "install", "--set", "profile=default", "-f", values_file, "-y"])
        print("Istio installed successfully.")
    except Exception as e:
        print(f"Error installing Istio: {e}")
        raise e
    # finally using the kubeconfi
    pods = get_pods_using_namespace("istio-system")
    for pod in pods.items:
        click.echo(f"Pod name: {pod.metadata.name}, Pod status: {pod.status.phase}")


def remove_domain_server():
    delete_resources_from_url(DOMAIN_SERVER_URL, "removing Knative Default Domain")


def uninstall_istio_from_cluster():
    """Uninstall Istio from the cluster."""
    # remove knative istion controller
    delete_resources_from_url(KNATIVE_ISTIO_CONTROLLER_URL, "uninstalling Knative Net Istio")
    # remove istio
    try:
        subprocess.run(["istioctl", "x", "uninstall", "--purge", "-y"])
        click.echo("Istio uninstalled successfully.")
    except Exception as e:
        click.echo(f"Error uninstalling Istio: {e}")


def install_net_istio():
    apply_yaml_from_url(KNATIVE_ISTIO_CONTROLLER_URL, "installing Knative Net Istio")


def install_default_domain():
    apply_yaml_from_url(DOMAIN_SERVER_URL, "installing Knative Default Domain")


def configure_ssl_for_ingress_gateway(certificate_arn: str, ssl_ports: str = "443") -> bool:
    """
    Configure SSL certificate for istio-ingressgateway service
    Args:
        certificate_arn: ACM certificate ARN
        ssl_ports: Comma-separated list of SSL ports (default: "443")
    Returns:
        bool: True if successful, False otherwise
    """
    try:
        # Load kubernetes config
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return False
        api_client = config.new_client_from_config(context=context_name)
        v1 = client.CoreV1Api(api_client)

        # Get existing service
        service = v1.read_namespaced_service(
            name='istio-ingressgateway',
            namespace='istio-system'
        )

        # Prepare annotations
        if not service.metadata.annotations:
            service.metadata.annotations = {}

        service.metadata.annotations.update({
            'service.beta.kubernetes.io/aws-load-balancer-ssl-cert': certificate_arn,
            'service.beta.kubernetes.io/aws-load-balancer-ssl-ports': ssl_ports,
            'service.beta.kubernetes.io/aws-load-balancer-backend-protocol': "http"
        })

        # Update service
        v1.patch_namespaced_service(
            name='istio-ingressgateway',
            namespace='istio-system',
            body=service
        )

        return True

    except client.exceptions.ApiException as e:
        click.echo(f"Failed to configure SSL: {str(e)}")
        return False
    except Exception as e:
        click.echo(f"Unexpected error configuring SSL: {str(e)}")
        return False

def configure_443_port_for_gateway() -> bool:
    """
    Add port 443 support to knative-ingress-gateway
    Returns:
        bool: True if successful, False otherwise
    """
    try:
        # Load kubernetes config
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return False
        api_client = config.new_client_from_config(context=context_name)
        custom_api = client.CustomObjectsApi(api_client)

        # Get existing gateway
        gateway = custom_api.get_namespaced_custom_object(
            group="networking.istio.io",
            version="v1",
            namespace="knative-serving",
            plural="gateways",
            name="knative-ingress-gateway"
        )

        # Check if 443 port is already configured
        for server in gateway['spec']['servers']:
            if server.get('port', {}).get('number') == 443:
                click.echo("Port 443 already configured in gateway")
                return True

        # Add new server for port 443
        gateway['spec']['servers'].append({
            'hosts': ['*'],
            'port': {
                'name': 'http-443',
                'number': 443,
                'protocol': 'HTTP'
            }
        })

        # Patch gateway
        custom_api.patch_namespaced_custom_object(
            group="networking.istio.io",
            version="v1",
            namespace="knative-serving",
            plural="gateways",
            name="knative-ingress-gateway",
            body=gateway
        )

        click.echo("Successfully added port 443 to gateway")
        return True

    except client.exceptions.ApiException as e:
        click.echo(f"Failed to configure gateway: {str(e)}")
        return False
    except Exception as e:
        click.echo(f"Unexpected error configuring gateway: {str(e)}")
        return False



from http.server import BaseHTTPRequestHandler, HTTPServer
import urllib.parse
import threading
import ssl
import warnings
from urllib3.exceptions import InsecureRequestWarning
from tensorkube.helpers import verify_user, save_user_credentials
import click

# Suppress only the specific InsecureRequestWarning
warnings.filterwarnings('ignore', category=InsecureRequestWarning)

class OAuthHandler(BaseHTTPRequestHandler):
    server_should_shutdown = False
    def log_message(self, format, *args):
        # Override to prevent any logs from being printed
        return
    
    def do_GET(self):
        # Parse the query string
        query = urllib.parse.urlparse(self.path).query
        params = urllib.parse.parse_qs(query)

        # Extract the token (you may need to adjust the key based on the OAuth provider)
        token = params.get('logged_in_token', [''])[0]
        session_id = params.get('session_id', [''])[0]

        # Respond to the browser
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

        # Save the token to a file
        user_email = verify_user(session_id=session_id, token=token)
        if user_email:
            self.wfile.write(b"Login successful! You can close this window now.")
            save_user_credentials(session_id=session_id, token=token)
            click.echo("Logged in successfully.")
        else:
            self.wfile.write(b"Login Failed! Please try again.")
            click.echo("Failed to log in. Error while verifying user credentials.")

        # Shut down the server
        OAuthHandler.server_should_shutdown = True

def run_server(server_class=HTTPServer, handler_class=OAuthHandler, port=8147):
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    # print(f"Starting server on port {port}...")

    try:
        while not handler_class.server_should_shutdown:
            httpd.handle_request()
    except KeyboardInterrupt:
        pass
    httpd.server_close()



import os
from typing import Optional, List

import boto3
import click
from tqdm import tqdm

from tensorkube.constants import DATASET_BUCKET_TYPE
from tensorkube.services.aws_service import get_bucket_name
from tensorkube.services.s3_service import check_if_a_bucket_type_exists, create_s3_bucket, list_bucket_contents, \
    S3ObjectInfo, upload_file, delete_file_from_s3


def ensure_dataset_bucket_exists() -> str:
    if check_if_a_bucket_type_exists(type=DATASET_BUCKET_TYPE):
        return get_bucket_name(type=DATASET_BUCKET_TYPE)
    ## doesnt exist create new bucket
    bucket_name = get_bucket_name(type=DATASET_BUCKET_TYPE)
    ## create bucket
    create_s3_bucket(bucket_name)
    return bucket_name


def list_datasets()->Optional[List[S3ObjectInfo]]:
    if not check_if_a_bucket_type_exists(type=DATASET_BUCKET_TYPE):
        click.echo(click.style("No datasets have ever been created.", fg='yellow'))
        return None

    datasets = list_bucket_contents(bucket_name=get_bucket_name(type=DATASET_BUCKET_TYPE))
    return datasets

def check_if_dataset_exists(dataset_id: str) -> bool:
    existing_datasets = list_datasets()
    # The name/path of the file in S3
    object_name = f"{dataset_id}.jsonl"
    if existing_datasets:
        if any(d.key == object_name for d in existing_datasets):
            return True
    return False


def upload_tensorkube_dataset(dataset_id: str, file_path: str) -> bool:
    """
    Uploads a dataset file to S3 with progress bar

    Args:
        dataset_id: Unique identifier for the dataset
        file_path: Path to the local JSONL file

    Returns:
        bool: True if upload was successful, False otherwise
    """
    try:
        # Ensure bucket exists
        bucket_name = ensure_dataset_bucket_exists()

        # The name/path of the file in S3
        object_name = f"{dataset_id}.jsonl"

        if check_if_dataset_exists(dataset_id):
            click.echo(click.style(f"Dataset with ID '{dataset_id}' already exists.", fg='red'))
            return False
        # Create s3 client
        s3_client = boto3.client('s3')
        # Get file size for progress bar
        file_size = os.path.getsize(file_path)

        # Create progress bar
        progress_bar = tqdm(
            total=file_size,
            unit='B',
            unit_scale=True,
            desc=f"Uploading {dataset_id}",
            leave=True
        )

        # Upload file
        upload_file(
            file_name=file_path,
            bucket_name=bucket_name,
            s3_key=object_name,  # This is the name/path in S3
            s3_client=s3_client,
            progress_bar=progress_bar
        )

        click.echo(click.style(f"Successfully uploaded dataset '{dataset_id}'", fg='green'))
        return True

    except Exception as e:
        print(f"Error uploading dataset: {e}")
        click.echo(click.style(f"Failed to upload dataset: {str(e)}", fg='red'))
        return False


def delete_tensorkube_dataset(dataset_id: str) -> bool:
    """
    Deletes a dataset from S3 using its ID

    Args:
        dataset_id: ID of the dataset to delete

    Returns:
        bool: True if deletion was successful, False otherwise
    """
    try:
        # Check if datasets bucket exists
        if not check_if_a_bucket_type_exists(type=DATASET_BUCKET_TYPE):
            click.echo(click.style("No datasets bucket exists.", fg='yellow'))
            return False

        bucket_name = get_bucket_name(type=DATASET_BUCKET_TYPE)
        object_name = f"{dataset_id}.jsonl"

        if not check_if_dataset_exists(dataset_id):
            click.echo(click.style(f"Dataset with ID '{dataset_id}' does not exists.", fg='red'))
            return False

        # Delete the dataset
        if delete_file_from_s3(bucket_name, object_name):
            click.echo(click.style(f"Successfully deleted dataset '{dataset_id}'", fg='green'))
            return True
        else:
            click.echo(click.style(f"Failed to delete dataset '{dataset_id}'", fg='red'))
            return False

    except Exception as e:
        print(f"Error deleting dataset: {e}")
        click.echo(click.style(f"Failed to delete dataset: {str(e)}", fg='red'))
        return False




import click
import boto3
import json
from pkg_resources import resource_filename
from tensorkube.constants import REGION, get_cluster_name


def get_or_create_ecr_repository(sanitised_project_name:str):
    ecr = boto3.client('ecr')
    repository_name = f"{get_cluster_name()}-{sanitised_project_name}"
    ecr_repo_lifecycle_policy = None
    try:
        #TODO! verify if this works. deletion takes upto 24 hours.
        ecr_repo_lifecycle_policy_config_file_path = resource_filename('tensorkube', 
                                                                       'configurations/aws_configs/ecr_repo_lifecycle_policy.json')
        with open(ecr_repo_lifecycle_policy_config_file_path, 'r') as f:
            ecr_repo_lifecycle_policy = json.dumps(json.load(f))
    except Exception as e:
        click.echo(f"Failed to read ECR repository lifecycle policy config file. Error: {e}")
        return None    
    try:
        response = ecr.create_repository(
            repositoryName=repository_name
        )
        ecr.put_lifecycle_policy(
            repositoryName=repository_name,
            lifecyclePolicyText=ecr_repo_lifecycle_policy
        )
        return response['repository']['repositoryUri']
    except ecr.exceptions.RepositoryAlreadyExistsException:
        click.echo(f"Repository '{repository_name}' already exists.")
        response = ecr.describe_repositories(repositoryNames=[repository_name])
        return response['repositories'][0]['repositoryUri']

def list_all_repositories(region: str=REGION):
    #TODO! pagination
    #NOTE: right now only 100 repositories are returned.
    ecr = boto3.client('ecr', region_name=region)
    response = ecr.describe_repositories()
    return response['repositories']


def delete_ecr_repository(repository_name:str, region:str=REGION):
    ecr = boto3.client('ecr', region_name=region)
    response = ecr.delete_repository(
        repositoryName=repository_name,
        force=True
    )
    return response


def delete_all_tensorkube_ecr_repositories():
    repositories = list_all_repositories()
    for repo in repositories:
        if get_cluster_name() in repo['repositoryName']:
            click.echo(f"Deleting repository '{repo['repositoryName']}'...")
            delete_ecr_repository(repo['repositoryName'])

def patch_lifecycle_all_tensorkube_ecr_repositories():
    ecr_repo_lifecycle_policy = None
    ecr = boto3.client('ecr')
    try:
        ecr_repo_lifecycle_policy_config_file_path = resource_filename('tensorkube', 
                                                                    'configurations/aws_configs/ecr_repo_lifecycle_policy.json')
        with open(ecr_repo_lifecycle_policy_config_file_path, 'r') as f:
            ecr_repo_lifecycle_policy = json.dumps(json.load(f))
    except Exception as e:
        click.echo(f"Failed to read ECR repository lifecycle policy config file. Error: {e}")
        return None    
    repositories = list_all_repositories()
    for repo in repositories:
        if get_cluster_name() in repo['repositoryName']:
            click.echo(f"Patching lifecycle policy for repository '{repo['repositoryName']}'...")
            ecr.put_lifecycle_policy(
                repositoryName=repo['repositoryName'],
                lifecyclePolicyText=ecr_repo_lifecycle_policy
            )
            click.echo(f"Lifecycle policy patched for repository '{repo['repositoryName']}'.")


class CLIVersionError(Exception):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.message = message
        self.code = code

    def __str__(self):
        if self.code:
            return f"[Error Code {self.code}]: {self.message}"
        return self.message


import json
import os

import boto3
import click
import yaml
from kubernetes import config, client
from pkg_resources import resource_filename
from typing import Optional, List

from tensorkube.constants import REGION, get_cluster_name, CliColors
from tensorkube.helpers import sanitise_name, extract_workdir_from_dockerfile, extract_command_from_dockerfile
from tensorkube.services.aws_service import get_aws_account_id, get_bucket_name, get_credentials
from tensorkube.services.dynamodb_service import get_dynamodb_table_name
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url, install_keda, delete_keda_from_cluster
from tensorkube.services.environment_service import delete_environment, create_new_environment
from tensorkube.services.filesystem_service import delete_efs_directory_for_deployment
from tensorkube.services.iam_service import create_sqs_access_policy, create_sqs_access_role, attach_role_policy, \
    get_role_name_for_prefix, create_dynamo_access_policy, delete_policy, detach_role_policy, delete_iam_role
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name, get_efs_claim_name, \
    list_keda_scaled_jobs, list_trigger_authentications, delete_trigger_authentication, create_service_account, \
    create_k8s_role, create_role_binding, create_aws_secret, is_existing_scaled_job, generate_kubeconfig
from tensorkube.services.knative_service import get_instance_family_from_gpu_type, get_supported_gpu_families
from tensorkube.services.s3_service import create_s3_bucket
from tensorkube.services.sqs_service import create_sqs_queue, queue_message, delete_sqs_queue
from tensorkube.services.nydus import get_nydus_runtime_class_name
from tensorkube.services.nodepool import get_gpu_nodepool



def create_cloud_resources_for_queued_job_support():
    cluster_name = get_cluster_name()
    oidc_issuer_url = get_cluster_oidc_issuer_url(cluster_name)

    sqs_policy_name = f"{cluster_name}-sqs-access-policy"
    role_name = f"{cluster_name}-sqs-access-role"
    policy = create_sqs_access_policy(sqs_policy_name)
    dyanmo_policy_name = f"{cluster_name}-dynamo-access-policy"
    create_dynamo_access_policy(dyanmo_policy_name)
    click.echo("Policy created")
    role = create_sqs_access_role(get_aws_account_id(), oidc_issuer_url, role_name, 'keda', 'keda-operator')
    click.echo("Role created")
    attach_role_policy(get_aws_account_id(), sqs_policy_name, role_name)
    attach_role_policy(get_aws_account_id(), dyanmo_policy_name, role_name)
    click.echo("Policy attached to role")

    # TODO!: on Nydus implementation, create new role and service account for combined access to SQS and DynamoDB to be used in ScaledJob
    eksctl_role = get_role_name_for_prefix(prefix=f"eksctl-{get_cluster_name()}-nodegroup-")
    attach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name, role_name=eksctl_role)
    karpenter_role = get_role_name_for_prefix(prefix=f"KarpenterNodeRole-{get_cluster_name()}")
    attach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name, role_name=karpenter_role)
    attach_role_policy(account_no=get_aws_account_id(), policy_name=dyanmo_policy_name, role_name=eksctl_role)
    attach_role_policy(account_no=get_aws_account_id(), policy_name=dyanmo_policy_name, role_name=karpenter_role)
    click.echo("SQS access policy attached to nodes")

    create_new_environment('keda')

    installed = install_keda(role['Role']['Arn'])
    if not installed:
        click.echo("Error installing Keda")
        return

    click.echo("Keda installed")
    create_trigger_authentication_for_aws_sqs(role['Role']['Arn'])
    click.echo("Trigger authentication created")
    # create train bucket for the keda environment
    bucket_name = get_bucket_name(env_name='keda', type='train')
    create_s3_bucket(bucket_name)
    create_aws_secret(get_credentials(),"keda")
    click.echo("S3 train bucket created for keda env")

    create_table_for_job_status(region=REGION)


def create_trigger_authentication_for_aws_sqs(sqs_access_iam_role_arn: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    trigger_auth_file_path = resource_filename('tensorkube', 'configurations/build_configs/queue_trigger_auth.yaml')
    with open(trigger_auth_file_path, 'r') as f:
        trigger_auth_json = yaml.safe_load(f)
    trigger_auth_json['spec']['podIdentity']['roleArn'] = sqs_access_iam_role_arn

    k8s_client.create_namespaced_custom_object('keda.sh', 'v1alpha1', 'keda', 'triggerauthentications',
                                               trigger_auth_json)


def get_queue_name_for_job(job_name: str):
    sanitised_job_name = sanitise_name(job_name)
    return f"{get_cluster_name()}-{sanitised_job_name}-queue"



def get_job_queue_url_for_job(job_name: str):
    queue_name = get_queue_name_for_job(job_name)
    sqs = boto3.client('sqs', region_name=REGION)
    response = sqs.get_queue_url(QueueName=queue_name)
    return response['QueueUrl']


def is_existing_queued_job(job_name: str):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    return is_existing_scaled_job(job_name=job_name, namespace="keda", context_name=context_name)


def deploy_job(job_name: str, gpus: int, gpu_type: Optional[str], cpu: int, memory: int, max_scale: int,
               env: str , image_tag:str, update: bool = False, region: str = REGION, cwd: Optional[str] =None,
               sanitised_project_name: Optional[str] = None, secrets: List[str] = [],
               context_name: Optional[str] = None, job_type: Optional[str] = None, enable_efs: bool = False):
    click.echo("Deploying job...")
    # Load kube config
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)
    if job_type == 'axolotl':
        scaled_job_file_path = resource_filename('tensorkube',
                                           'configurations/build_configs/fine_tuning_scaled_job.yaml')
        with open(scaled_job_file_path, 'r') as f:
            scaled_job_yaml = f.read()

        scaled_job_yaml = scaled_job_yaml.replace('${IMAGE_TAG}', image_tag)
        scaled_job_yaml = scaled_job_yaml.replace('${GPUS}', str(gpus))
        scaled_job_yaml = scaled_job_yaml.replace('${GPU_TYPE}', gpu_type)
        scaled_job_json = yaml.safe_load(scaled_job_yaml)
        job_queue_url = create_sqs_queue(get_queue_name_for_job(job_name))

        bucket_name = get_bucket_name(env_name=env, type='train')
        scaled_job_json['metadata']['name'] = job_name
        scaled_job_json['spec']['triggers'][0]['metadata']['queueURL'] = job_queue_url
        scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['env'].extend([
            {'name': 'AWS_REGION', 'value': region}, {'name': 'QUEUE_URL', 'value': job_queue_url}, {'name': 'JOB_NAME', 'value': job_name}, {'name': 'LORA_ADAPTER_BUCKET', 'value': bucket_name}])
        scaled_job_json['spec']['triggers'][0]['metadata']['region'] = region
        scaled_job_json['spec']['maxReplicaCount'] = max_scale

        if secrets:
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['envFrom'] = [{
                'secretRef': {
                    'name': secret_name
                }
            } for secret_name in secrets]
                
        scaled_job_json['spec']['jobTargetRef']['template']['spec']['nodeSelector'] = {
            'karpenter.k8s.aws/instance-family': get_instance_family_from_gpu_type(gpu_type),
        }
    else:
        scaled_job_file_path = resource_filename('tensorkube', 'configurations/build_configs/scaled_job.yaml')
        with open(scaled_job_file_path, 'r') as f:
            scaled_job_json = yaml.safe_load(f)

        job_queue_url = create_sqs_queue(get_queue_name_for_job(job_name))

        scaled_job_json['metadata']['name'] = job_name
        scaled_job_json['spec']['triggers'][0]['metadata']['queueURL'] = job_queue_url
        scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['env'] = [
            {'name': 'AWS_REGION', 'value': region}, {'name': 'QUEUE_URL', 'value': job_queue_url},
            {'name': 'JOB_NAME', 'value': job_name}]
        scaled_job_json['spec']['triggers'][0]['metadata']['region'] = region
        scaled_job_json['spec']['maxReplicaCount'] = max_scale

        scaled_job_json['spec']['jobTargetRef']['template']['spec']['initContainers'][0]['env'] = [
            {'name': 'AWS_REGION', 'value': region}]

        scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources'] = {'requests': {},
                                                                                                     'limits': {}}
        if gpus > 0:
            if memory > 2000:
                scaled_job_json['spec']['template']['spec']['containers'][0]['resources']['requests'][
                    'memory'] = f'{str(int(memory))}M'
            if cpu > 2000:
                scaled_job_json['spec']['template']['spec']['containers'][0]['resources']['requests'][
                    'cpu'] = f'{str(int(cpu))}m'
    
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['requests'][
                'nvidia.com/gpu'] = gpus
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['limits'][
                'nvidia.com/gpu'] = gpus

            scaled_job_json['spec']['jobTargetRef']['template']['spec']['nodeSelector'] = {
                'karpenter.k8s.aws/instance-family': get_instance_family_from_gpu_type(gpu_type)}
        else:
    
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['requests'][
                'cpu'] = f'{str(int(cpu))}m'
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['requests'][
                'memory'] = f'{str(int(memory))}M'
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['limits'][
                'cpu'] = f'{str(int(cpu))}m'
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['resources']['limits'][
                'memory'] = f'{str(int(memory))}M'

            scaled_job_json['spec']['jobTargetRef']['template']['spec']['affinity'] = {
                "nodeAffinity": {
                    "requiredDuringSchedulingIgnoredDuringExecution": {
                    "nodeSelectorTerms": [ {
                        "matchExpressions": [ {
                            "key": "karpenter.k8s.aws/instance-family",
                            "operator": "NotIn",
                            "values": get_supported_gpu_families()
                        } ]
                    } ]
            }}}

        if enable_efs:
            dockerfile_path = cwd + "/Dockerfile"
            workdir = extract_workdir_from_dockerfile(dockerfile_path)
            command = extract_command_from_dockerfile(dockerfile_path)
            if workdir:
                final_command = f"cd {workdir} && {command}"
            else:
                final_command = command

            if secrets:
                secrets_to_env_vars_command = """\
                    folder_path="/mnt/secrets"

                    # Initialize an empty string to hold the environment variables
                    env_vars=""

                    # Loop through each file in the folder
                    for file in "$folder_path"/*; do
                        if [[ -f $file ]]; then
                            # Get the filename without the path
                            filename=$(basename "$file")

                            # Get the contents of the file
                            contents=$(<"$file")

                            # Escape the contents to handle any special characters or spaces
                            escaped_contents=$(printf '%q' "$contents")

                            # Append to the env_vars string in the format --env filename=contents
                            env_vars="$env_vars --env $filename=$escaped_contents"
                        fi
                    done"""
            else:
                secrets_to_env_vars_command = "env_vars=''"

            if gpus > 0:
                config_nvidia_ctk_commands = """sudo nvidia-ctk cdi generate --output=/etc/cdi/nvidia.yaml
                           nvidia-ctk cdi list
                        """
                podman_gpu_tags = "--gpus all --env NVIDIA_VISIBLE_DEVICES=all --env NVIDIA_DRIVER_CAPABILITIES=compute,utility"
        
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0][
                    'image'] = "tensorfuse/podman-nvidia:v1"
                
            else:
                config_nvidia_ctk_commands = ""
                podman_gpu_tags = ""
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0][
                'image'] = "quay.io/podman/stable"
            
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['command'] = [
                "/bin/sh", "-c",f"""{config_nvidia_ctk_commands}
                {secrets_to_env_vars_command}
                sed -i 's|mount_program = "/usr/bin/fuse-overlayfs"|mount_program = ""|' /etc/containers/storage.conf
                sudo podman run --name mycontainer $env_vars  --env QUEUE_URL={job_queue_url} --env AWS_REGION={region} --env JOB_NAME={job_name} \
                -v /mnt/shared:/mnt/shared \
                {podman_gpu_tags} --network=host \
                --rootfs /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs:O sh -c "{final_command}" """]
            
            
            for volume in scaled_job_json['spec']['jobTargetRef']['template']['spec']['volumes']:
                if volume['name'] == 'efs-storage':
                    volume['persistentVolumeClaim']['claimName'] = get_efs_claim_name(env_name=env)
            if secrets:
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['volumes'].append({'name': 'secrets',
                    'projected': {'sources': [{'secret': {'name': secret_name}} for secret_name in secrets]}})
        
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['volumeMounts'].append(
                    {'name': 'secrets', 'mountPath': '/mnt/secrets', 'readOnly': True})
                
        else:
            # Remove the EFS volume
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['volumes'] = [volume for volume in scaled_job_json['spec']['jobTargetRef']['template']['spec']['volumes'] if volume['name'] != 'efs-storage']
            # Remove the volume mount
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['volumeMounts'] = [volume for volume in scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['volumeMounts'] if volume['name'] != 'efs-storage']
    
            if secrets:  
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['envFrom'] = [{
                    'secretRef': {
                        'name': secret_name
                    }
                } for secret_name in secrets]
            
            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['env'] = [
                {'name': 'AWS_REGION', 'value': region},
                {'name': 'QUEUE_URL', 'value': job_queue_url},
                {'name': 'JOB_NAME', 'value': job_name},
            ]

            scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['image'] = image_tag
            del scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['securityContext']
            del scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][0]['command']
            if gpus > 0:
                scaled_job_json['spec']['jobTargetRef']['template']['spec']['runtimeClassName'] = get_nydus_runtime_class_name()
                if 'nodeSelector' in scaled_job_json['spec']['jobTargetRef']['template']['spec']:
                    scaled_job_json['spec']['jobTargetRef']['template']['spec']['nodeSelector']['nodepool'] = get_gpu_nodepool()
                else:
                    scaled_job_json['spec']['jobTargetRef']['template']['spec']['nodeSelector'] = {'nodepool': get_gpu_nodepool()}
            
    
    scaled_job_json['spec']['jobTargetRef']['template']['spec']['initContainers'][0]['command'] = ["/bin/sh", "-c", f"""\
            export QUEUE_URL={job_queue_url}
            export TABLE_NAME={get_dynamodb_table_name()}
            export JOB_NAME={job_name}
            export MSG=$(aws sqs receive-message --queue-url $QUEUE_URL --max-number-of-messages 1 --query 'Messages[0]')
            if [ "$MSG" == "null" ]; then
              echo "No message in queue"
            else
              echo $MSG | jq -r '.Body' | jq -r '.job_id' > /mnt/shared/sqs_message_id.txt
              echo $MSG | jq -r '.Body' | jq -r '.job_payload' > /mnt/shared/sqs_message_payload.txt
              echo $MSG | jq -r '.ReceiptHandle' > /mnt/shared/receipt_handle.txt
              echo "Message received with job_id: $(cat /mnt/shared/sqs_message_id.txt)" 
              aws dynamodb put-item --table-name $TABLE_NAME --item '{{
                "job_name": {{"S": "'"$JOB_NAME"'"}},
                "job_id": {{"S": "'"$(cat /mnt/shared/sqs_message_id.txt)"'"}},
                "status": {{"S": "PROCESSING"}}
              }}'
            fi
            """]
    
    scaled_job_json['spec']['jobTargetRef']['template']['spec'][
        'serviceAccountName'] = get_job_sidecar_service_account_name()
    scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][1]['env'] = [
        {'name': 'AWS_REGION', 'value': region}]
    scaled_job_json['spec']['jobTargetRef']['template']['spec']['containers'][1]['command'] = ["/bin/sh", "-c", f"""\
            export QUEUE_URL={job_queue_url}
            export TABLE_NAME={get_dynamodb_table_name()}
            export JOB_NAME={job_name}
            echo "Tracking job status"
            while true; do
                status=$(kubectl get pod ${{HOSTNAME}} -n keda -o jsonpath='{{.status.containerStatuses[?(@.name=="sqs-job-1")].state.terminated}}')
                if [ -n "$status" ]; then
                    echo "Main container finished."
                    aws sqs delete-message --queue-url $QUEUE_URL --receipt-handle $(cat /mnt/shared/receipt_handle.txt) --region $AWS_REGION
                    
                    exit_code=$(kubectl get pod ${{HOSTNAME}} -n keda -o jsonpath='{{.status.containerStatuses[?(@.name=="sqs-job-1")].state.terminated.exitCode}}')
                    if [ "$exit_code" -eq 0 ]; then
                        echo "Main container succeeded. Running post-completion command."
                        aws dynamodb put-item --table-name $TABLE_NAME --item '{{ "job_name": {{"S": "'"$JOB_NAME"'"}},
                            "job_id": {{"S": "'"$(cat /mnt/shared/sqs_message_id.txt)"'"}}, "status": {{"S": "SUCCESS"}} }}' --region $AWS_REGION
                    else
                        echo "Main container failed. Running post-error command."
                        aws dynamodb put-item --table-name $TABLE_NAME --item '{{ "job_name": {{"S": "'"$JOB_NAME"'"}},
                            "job_id": {{"S": "'"$(cat /mnt/shared/sqs_message_id.txt)"'"}}, "status": {{"S": "ERROR"}} }}' --region $AWS_REGION
                    fi
                    break
                    fi
                    sleep 5
                done
        """]


    if update:
        try:
            group = "keda.sh"
            version = "v1alpha1"
            namespace = "keda"
            plural = "scaledjobs"
            existing_job = k8s_client.get_namespaced_custom_object(group, version, namespace, plural, job_name)
            scaled_job_json['metadata']['resourceVersion'] = existing_job['metadata']['resourceVersion']
            k8s_client.replace_namespaced_custom_object(group, version, namespace, plural, job_name, scaled_job_json)
            click.echo("Job updated successfully.")
        except Exception as e:
            print(f"Error while updating job: {e}")
    else:
        try:
            group = "keda.sh"
            version = "v1alpha1"
            namespace = "keda"
            plural = "scaledjobs"
            k8s_client.create_namespaced_custom_object(group, version, namespace, plural, scaled_job_json)
            click.echo("Job deployed successfully.")
            click.echo(f"Run the job by running: tensorkube job queue --job-name {job_name} --job-id <job_id> --payload <payload>")
        except Exception as e:
            print(f"Error while deploying job: {e}")

def queue_job(job_name: str, job_id: str, job_payload: str, region: str = REGION):
    dynamodb = boto3.resource('dynamodb', region_name=region)
    table_name = f"{get_cluster_name()}-job-status"
    table = dynamodb.Table(table_name)

    # Check if the job already exists
    try:
        response = table.get_item(Key={'job_name': job_name, 'job_id': job_id})
        if 'Item' in response:
            click.echo(click.style(f"Job {job_name} with ID {job_id} already exists. Skipping queue.", fg=CliColors.ERROR.value))
            return False
    except Exception as e:
        print(f"Error checking job existence: {e}")
        raise e
    queur_url = get_job_queue_url_for_job(job_name)
    msg = {"job_id": job_id, "job_payload": job_payload}
    msg_str = json.dumps(msg)
    queue_message(queur_url, msg_str, region=REGION)
    set_job_status(job_name, job_id, status="QUEUED")
    return True


def delete_job(job_name: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None

    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    try:
        group = "keda.sh"
        version = "v1alpha1"
        namespace = "keda"
        plural = "scaledjobs"
        k8s_client.delete_namespaced_custom_object(group, version, namespace, plural, job_name)
        click.echo(f"Job {job_name} deleted successfully.")
    except Exception as e:
        print(f"Error while deleting job: {e}")


def delete_all_job_resources(job_name: str):
    delete_job(job_name=job_name)
    queue_name = get_queue_name_for_job(job_name)
    delete_sqs_queue(queue_name)
    delete_job_in_dynamo(job_name)
    delete_efs_directory_for_deployment(sanitise_name(job_name), 'keda')


def delete_job_in_dynamo(job_name: str):
    dynamodb = boto3.resource('dynamodb', region_name=REGION)
    table = dynamodb.Table(get_dynamodb_table_name())
     # Query items with the specific partition key
    response = table.query(
        KeyConditionExpression=boto3.dynamodb.conditions.Key("job_name").eq(job_name)
    )
    #TODO: Handle pagination
    items = response.get('Items', [])
    with table.batch_writer() as batch:
        for item in items:
            batch.delete_item(
                Key={"job_name": item["job_name"], "job_id": item["job_id"]}
            )
    
    click.echo(f"Job {job_name} deleted from DynamoDB.")

def delete_dynamo_table():
    dynamodb = boto3.resource('dynamodb', region_name=REGION)
    table_name = get_dynamodb_table_name()
    try:
        table = dynamodb.Table(table_name)
        table.delete()
        print(f"Table {table_name} deleted successfully.")
    except Exception as e:
        print(f"Error deleting table: {e}")


def teardown_job_queue_support():
    jobs = list_keda_scaled_jobs()
    for job in jobs['items']:
        job_name = job['metadata']['name']
        delete_all_job_resources(job_name)

    trigger_auths = list_trigger_authentications()
    for trigger_auth in trigger_auths['items']:
        trigger_auth_name = trigger_auth['metadata']['name']
        delete_trigger_authentication(trigger_auth_name)
    delete_dynamo_table()
    cluster_name = get_cluster_name()
    sqs_policy_name = f"{cluster_name}-sqs-access-policy"
    sqs_role_name = f"{cluster_name}-sqs-access-role"
    dynamo_policy_name = f"{cluster_name}-dynamo-access-policy"

    eksctl_role = get_role_name_for_prefix(prefix=f"eksctl-{get_cluster_name()}-nodegroup-")
    karpenter_role = get_role_name_for_prefix(prefix=f"KarpenterNodeRole-{get_cluster_name()}")

    detach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name, role_name=eksctl_role)
    detach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name, role_name=karpenter_role)
    detach_role_policy(account_no=get_aws_account_id(), policy_name=dynamo_policy_name, role_name=eksctl_role)
    detach_role_policy(account_no=get_aws_account_id(), policy_name=dynamo_policy_name, role_name=karpenter_role)

    detach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name, role_name=sqs_role_name)
    delete_policy(account_no=get_aws_account_id(), policy_name=sqs_policy_name)
    delete_iam_role(role_name=sqs_role_name)

    delete_policy(account_no=get_aws_account_id(), policy_name=dynamo_policy_name)

    delete_keda_from_cluster()
    delete_environment('keda')


# EXPOSE
def get_queued_message():
    payload_file_path = "/mnt/shared/sqs_message_payload.txt"
    if os.path.exists(payload_file_path):
        with open(payload_file_path, 'r') as file:
            return file.read()
    else:
        raise Exception("Payload file not found.")

#EXPOSE
def queue_new_job(job_name: str, job_id: str, job_payload: str):
    try:
        generate_kubeconfig()
    except Exception as e:
        raise Exception(f"Failed to generate kubeconfig, {e}")

    exists, job = is_existing_queued_job(job_name)
    if not exists:
        click.echo(click.style(f"Job {job_name} does not exist. Please deploy the job first.", fg=CliColors.ERROR.value))
        return False
    return queue_job(job_name, job_id, job_payload)


def delete_queued_message(receipt_handle: str):
    sqs = boto3.client('sqs', region_name=REGION)
    queue_url = os.getenv('QUEUE_URL', None)
    if not queue_url:
        raise Exception("QUEUE_URL not set in environment variables.")

    response = sqs.delete_message(QueueUrl=queue_url, ReceiptHandle=receipt_handle)
    return response


def create_table_for_job_status(region: str = REGION):
    table_name = get_dynamodb_table_name()
    dynamodb = boto3.resource('dynamodb', region_name=region)
    try:
        # Create the DynamoDB table
        table = dynamodb.create_table(TableName=table_name,
            KeySchema=[{'AttributeName': 'job_name', 'KeyType': 'HASH'  # Partition key
            }, {'AttributeName': 'job_id', 'KeyType': 'RANGE'  # Sort key
            }], AttributeDefinitions=[{'AttributeName': 'job_name', 'AttributeType': 'S'},
                {'AttributeName': 'job_id', 'AttributeType': 'S'}], BillingMode='PAY_PER_REQUEST'
            # Set BillingMode to On-Demand
        )

        # Wait until the table exists
        table.meta.client.get_waiter('table_exists').wait(TableName=table_name)
        print(f"Table {table_name} created successfully.")

    except Exception as e:
        print(f"Error creating table: {e}")


# EXPOSE
def set_job_status(job_name: str, job_id: str, status: str):
    dynamodb = boto3.resource('dynamodb', region_name=REGION)
    table = dynamodb.Table(get_dynamodb_table_name())
    response = table.put_item(Item={'job_name': job_name, 'job_id': job_id, 'status': status})
    return response

#EXPOSE
def get_job_status(job_name: str, job_id: str):
    dynamodb = boto3.resource('dynamodb', region_name=REGION)
    table = dynamodb.Table(get_dynamodb_table_name())
    response = table.get_item(Key={'job_name': job_name, 'job_id': job_id})
    return response['Item']['status'] if 'Item' in response else None


def get_job_sidecar_service_account_name(namespace: Optional[str] = None):
    if not namespace:
        namespace = "keda"
    return f"job-queue-sidecar-{namespace}-sa"


def create_sa_role_rb_for_job_sidecar(namespace: Optional[str] = None):
    if not namespace:
        namespace = "keda"
    sa_name = get_job_sidecar_service_account_name(namespace)
    role_name = f"job-queue-sidecar-{namespace}-role"
    rb_name = f"job-queue-sidecar-{namespace}-rb"

    pod_tracking_rules = [{"apiGroups": [""], "resources": ["pods"], "verbs": ["get", "list"]}]

    create_service_account(name=sa_name, namespace=namespace)
    click.echo("Service Account created")
    create_k8s_role(role_name=role_name, rules=pod_tracking_rules, namespace=namespace)
    click.echo("Role created")
    create_role_binding(rb_name=rb_name, role_name=role_name, sa_name=sa_name, rb_namespace=namespace,
                        sa_namespace=namespace)
    click.echo("Role Binding created")



import os
import platform
import shutil
import subprocess
import uuid
from typing import Optional

import boto3
import botocore
import click
import semver

from tensorkube.constants import CliColors
from tensorkube.constants import get_cluster_name, DEFAULT_NAMESPACE, LOCKED_AWS_CLI_VERSION
from tensorkube.services.error import CLIVersionError
from tensorkube.services.s3_service import list_s3_buckets


def get_session_region():
    return boto3.session.Session().region_name


def get_eks_client():
    return boto3.client("eks")


def get_cloudformation_client():
    return boto3.client("cloudformation")


def get_ec2_client():
    return boto3.client("ec2")


def get_iam_client():
    return boto3.client("iam")


def get_aws_account_id():
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    return identity['Account']


def get_aws_user_arn() -> str:
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    return identity['Arn']


def get_principal_arn_from_identity_center_arn(arn: str) -> str:
    arn_parts = arn.split(':')
    account_no = arn_parts[4]
    role_name = arn_parts[5].split('/')[1]
    principal_arn = f"arn:aws:iam::{account_no}:role/aws-reserved/sso.amazonaws.com/{role_name}"
    return principal_arn


def get_aws_user_principal_arn() -> str:
    sts = boto3.client('sts')
    identity = sts.get_caller_identity()
    if 'assumed-role' in identity['Arn']:
        return get_principal_arn_from_identity_center_arn(identity['Arn'])
    return identity['Arn']


def get_karpenter_namespace():
    return "kube-system"


def get_karpenter_version():
    return "0.37.0"


def get_aws_default_region():
    return "us-east-1"


def get_kubernetes_context_name():
    return f"{get_cluster_name()}.{get_aws_default_region()}.eksctl.io"


def get_credentials():
    return boto3.Session().get_credentials().get_frozen_credentials()


def are_credentials_valid(credentials):
    sts = boto3.client('sts', aws_access_key_id=credentials.access_key, aws_secret_access_key=credentials.secret_key,
                       aws_session_token=credentials.token)
    try:
        sts.get_caller_identity()
        return True
    except botocore.exceptions.ClientError as e:
        return False


# TODO!: make function generic to get any config value
def get_bucket_name(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return bucket['Name']
    else:
        bucket_name = f"{prefix}{str(uuid.uuid4())[:18]}"
        if len(bucket_name) > 63:
            raise ValueError("Environment name is too long. Please use a shorter name.")
        return bucket_name


def check_and_install_aws_cli():
    """Check if aws cli is installed and if not install it."""
    try:
        result = subprocess.run(["aws", "--version"], capture_output=True, text=True, check=True)

        # The version is in the format: "aws-cli/2.13.4 Python/3.9.6 Linux/x86_64"
        version_output = result.stdout.strip()
        # Extract the AWS CLI version using string manipulation
        aws_version = semver.VersionInfo.parse(version_output.split(" ")[0].split("/")[1])
        locked_cli_version = semver.VersionInfo.parse(LOCKED_AWS_CLI_VERSION)
        if (aws_version.major, aws_version.minor) < (locked_cli_version.major, locked_cli_version.minor):
            text = f"AWS CLI version is {aws_version}. Please upgrade AWS CLI to version above {LOCKED_AWS_CLI_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("aws-cli not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        try:
            system = platform.system().lower()
            architecture = platform.machine()
            print(f"Detected operating system: {system.capitalize()}")
            print(f"Detected architecture: {architecture}")
            url = ""
            file_name = ""
            # Define the download URL and file name based on the OS
            if system == "linux":
                url = f"https://awscli.amazonaws.com/awscli-exe-linux-{architecture}-{LOCKED_AWS_CLI_VERSION}.zip"
                file_name = f"awscliv2-{LOCKED_AWS_CLI_VERSION}.zip"
            elif system == "darwin":
                url = f"https://awscli.amazonaws.com/AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
                file_name = f"AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
            else:
                print("Unsupported operating system. Please install AWS CLI manually.")
                raise Exception("Unsupported operating system.")
            # Download the installer
            print(f"Downloading AWS CLI installer from {url}...")
            subprocess.run(["curl", "-fsSL", url, "-o", file_name], check=True)
            print("AWS CLI installer downloaded successfully.")
            if system == "linux":
                # Unzip the installer
                print("Unzipping the AWS CLI installer...")
                subprocess.run(["unzip", "-o", file_name], check=True)
                print("AWS CLI installer unzipped successfully.")

                # Install the AWS CLI
                print("Installing AWS CLI...")
                subprocess.run(["sudo", "./aws/install"], check=True)
                print("AWS CLI installed successfully.")

            elif system == "darwin":
                # Use the macOS installer
                print("Installing AWS CLI using macOS package...")
                subprocess.run(["sudo", "installer", "-pkg", file_name, "-target", "/"], check=True)
                print("AWS CLI installed successfully.")
        except subprocess.CalledProcessError as e:
            print(f"Error occurred during installation: {e}")
        finally:
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists("aws"):
                shutil.rmtree("aws")
            print("Clean-up completed.")



import subprocess
import time
from threading import Thread
from typing import List, Optional, Dict, Any

import click
import yaml
from kubernetes import client, config, utils, watch
from kubernetes.client import ApiTypeError, ApiException
from pkg_resources import resource_filename

from tensorkube.constants import NAMESPACE, REGION, PodStatus, BUILD_TOOL, DEFAULT_NAMESPACE, TENSORFUSE_NAMESPACES, \
    get_cluster_name, CliColors
from tensorkube.services.aws_service import get_aws_account_id
import boto3
import os

def get_s3_pv_name(env_name: Optional[str] = None):
    if env_name:
        return f"s3-pv-env-{env_name}"
    return "s3-pv"


def get_s3_claim_name(env_name: Optional[str] = None):
    if env_name:
        return f"s3-claim-env-{env_name}"
    return "s3-claim"


def get_efs_claim_name(env_name: Optional[str] = None):
    if env_name:
        return f"efs-pvc-env-{env_name}"
    return "efs-pvc"


def create_namespace(namespace_name, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    namespace = client.V1Namespace()
    namespace.metadata = client.V1ObjectMeta(name=namespace_name)
    v1 = client.CoreV1Api(k8s_api_client)
    v1.create_namespace(body=namespace)


def create_docker_registry_secret(secret_name: str, namespace: str, base64_encoded_dockerconfigjson: str,
                                  context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)

    secret = client.V1Secret()
    secret.api_version = "v1"
    secret.kind = "Secret"
    secret.metadata = client.V1ObjectMeta(name=secret_name, namespace=namespace)
    secret.type = "kubernetes.io/dockerconfigjson"
    secret.data = {".dockerconfigjson": base64_encoded_dockerconfigjson}

    v1.create_namespaced_secret(namespace=namespace, body=secret)


def create_aws_secret(credentials, namespace: str = DEFAULT_NAMESPACE, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    secret_name = "aws-secret"

    secret = client.V1Secret()
    secret.metadata = client.V1ObjectMeta(name=secret_name)
    secret.string_data = {"AWS_ACCESS_KEY_ID": credentials.access_key, "AWS_SECRET_ACCESS_KEY": credentials.secret_key,
                          "AWS_SESSION_TOKEN": credentials.token}

    try:
        # Check if the secret already exists
        existing_secret = v1.read_namespaced_secret(name=secret_name, namespace=namespace)
        # If the secret exists, update it
        v1.replace_namespaced_secret(name=secret_name, namespace=namespace, body=secret)
        print(f"Secret {secret_name} updated successfully in namespace {namespace}.")
    except ApiException as e:
        if e.status == 404:
            # Secret does not exist, create it
            v1.create_namespaced_secret(namespace=namespace, body=secret)
            print(f"Secret {secret_name} created successfully in namespace {namespace}.")
        else:
            print(f"An error occurred: {e}")
            raise e


def delete_aws_secret(namespace: str = DEFAULT_NAMESPACE, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    try:
        v1.read_namespaced_secret(name="aws-secret", namespace=namespace)
    except client.ApiException as e:
        if e.status == 404:
            return
        else:
            raise
    v1.delete_namespaced_secret(name="aws-secret", namespace=namespace)


def create_build_pv_and_pvc(bucket_name: str, region: str = REGION, env: Optional[str] = None,
                            context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    pv_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/pv.yaml')
    pvc_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/pvc.yaml')
    with open(pv_config_file_path) as f:
        pv = yaml.safe_load(f)
    with open(pvc_config_file_path) as f:
        pvc = yaml.safe_load(f)

    env_namespace = env if env else DEFAULT_NAMESPACE

    pv['spec']['mountOptions'] = ["allow-delete", "region {}".format(region)]
    pv['spec']['csi']['volumeAttributes']['bucketName'] = bucket_name
    pv['metadata']['namespace'] = env_namespace
    pv['metadata']['name'] = get_s3_pv_name(env_name=env)

    pvc['metadata']['namespace'] = env_namespace
    pvc['metadata']['name'] = get_s3_claim_name(env_name=env)
    pvc['spec']['volumeName'] = get_s3_pv_name(env_name=env)

    v1 = client.CoreV1Api(k8s_api_client)

    pv_name = pv['metadata']['name']
    pvc_name = pvc['metadata']['name']

    try:
        # Check if the PV already exists
        v1.read_persistent_volume(name=pv_name)
        click.echo(f"PersistentVolume {pv_name} already exists. Skipping creation.")
    except ApiException as e:
        if e.status == 404:
            utils.create_from_dict(k8s_api_client, pv)
            click.echo(f"PersistentVolume {pv_name} created successfully.")
        else:
            click.echo(f"An error occurred while checking PersistentVolume: {e}")
            raise e

    try:
        # Check if the PVC already exists
        v1.read_namespaced_persistent_volume_claim(name=pvc_name, namespace=env_namespace)
        click.echo(f"PersistentVolumeClaim {pvc_name} already exists in namespace {env_namespace}. Skipping creation.")
    except ApiException as e:
        if e.status == 404:
            # PVC does not exist, proceed to create
            utils.create_from_dict(k8s_api_client, pvc)
            click.echo(f"PersistentVolumeClaim {pvc_name} created successfully in namespace {env_namespace}.")
        else:
            click.echo(f"An error occurred while checking PersistentVolumeClaim: {e}")
            raise e


        




def apply_image_cleanup_job(sanitised_project_name: str, image_tags: List[str], env: Optional[str] = None,
                            context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    cleanup_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/efs_cleanup_pod.yaml')
    with open(cleanup_config_file_path) as f:
        cleanup_config = yaml.safe_load(f)
    cleanup_config['metadata']['name'] = 'cleanup-{}'.format(sanitised_project_name)
    namespace_to_use = env if env else DEFAULT_NAMESPACE  # Replace 'default' with your default namespace if needed
    cleanup_config['metadata']['namespace'] = namespace_to_use

    for volume in cleanup_config['spec']['template']['spec']['volumes']:
        if volume['name'] == 'efs-pvc':
            volume['persistentVolumeClaim']['claimName'] = get_efs_claim_name(env_name=env)

    cleanup_config['spec']['template']['spec']['containers'][0]['command'] = ["/bin/sh", "-c",
                                                                              f"""cd /mnt/efs/images/{sanitised_project_name}
        echo 'Deleting all images except' {", ".join(image_tags)}
        find . -mindepth 1 -maxdepth 1 -type d ! -name {" ! -name ".join(image_tags)} -exec rm -rf {{}} +
        echo 'Deletion completed' """]

    utils.create_from_dict(k8s_api_client, cleanup_config)
    click.echo('Deployed a delete config job')


def get_build_job_pod_name(sanitised_project_name: str, namespace: str = NAMESPACE, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    pods = v1.list_namespaced_pod(namespace=namespace)
    for pod in pods.items:
        if pod.metadata.name.startswith("{}-{}".format(BUILD_TOOL, sanitised_project_name)):
            return pod.metadata.name
    return None


def check_pod_status(pod_name, namespace, context_name: Optional[str] = None):
    # Load kube config
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create a Kubernetes API client
    v1 = client.CoreV1Api(k8s_api_client)

    # Get the status of the pod
    pod_status = v1.read_namespaced_pod_status(name=pod_name, namespace=namespace)

    # Return the status of the pod
    return pod_status.status.phase


def find_and_delete_old_job(job_name: str, namespace: str = DEFAULT_NAMESPACE, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    job_pod_name = get_pod_name_corresponing_to_job(job_name, namespace)
    if job_pod_name:
        click.echo("Terminating pod {} corresponding to job {}".format(job_pod_name, job_name))
        v1.delete_namespaced_pod(name=job_pod_name, namespace=namespace)

    v1 = client.BatchV1Api(k8s_api_client)
    jobs = v1.list_namespaced_job(namespace=namespace)
    for job in jobs.items:
        if job.metadata.name == job_name:
            v1.delete_namespaced_job(name=job.metadata.name, namespace=namespace)
            return True

    return True


def delete_all_jobs_in_namespace(namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.BatchV1Api(k8s_api_client)
    jobs = v1.list_namespaced_job(namespace=namespace)
    for job in jobs.items:
        v1.delete_namespaced_job(name=job.metadata.name, namespace=namespace)
        print(f"Job {job.metadata.name} deletion initiated.")

        # Wait for the job to be deleted
        while True:
            try:
                v1.read_namespaced_job(name=job.metadata.name, namespace=namespace)
                time.sleep(1)  # Wait for 1 second before checking again
            except ApiException as e:
                if e.status == 404:
                    print(f"Job {job.metadata.name} deleted successfully.")
                    break
                else:
                    print(f"Error while waiting for deletion of job {job.metadata.name}: {e}")
                    raise e
    return True


def start_streaming_pod(pod_name, namespace, status=None, container_name=None, retry_number: int = 0,
                        max_retries: int = 4, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    # Create a stream to the pod
    # Initialize the Watch class
    watch_client = watch.Watch()
    # Stream events until the pod is ready
    print(f"Streaming events for pod {pod_name} in namespace {namespace}")
    try:
        events_streamed_upto_index = 0

        while True:
            pod = v1.read_namespaced_pod(name=pod_name, namespace=namespace)
            events = v1.list_namespaced_event(namespace=namespace, field_selector=f'involvedObject.name={pod_name}')
            for event in events.items[events_streamed_upto_index:]:
                print("Event: %s %s" % (event.type, event.message))
            events_streamed_upto_index = len(events.items)
            if pod.status.phase != 'Pending':
                break
            time.sleep(1)

    except KeyboardInterrupt:
        print("Log streaming stopped by user")
        return

    print(f"Streaming logs for pod {pod_name} in namespace {namespace}")

    try:
        last_log_printed = None
        while True:
            logs = v1.read_namespaced_pod_log(name=pod_name, namespace=namespace, container=container_name,
                                              since_seconds=5)
            logs_by_line = logs.split('\n')
            logs_by_line = logs_by_line[:len(logs_by_line) - 1]
            if logs_by_line:
                if not last_log_printed:
                    for log in logs_by_line:
                        print(log)
                else:
                    found_last_printed_log = False
                    for log in logs_by_line:
                        if found_last_printed_log:
                            print(log)
                            continue
                        if log == last_log_printed:
                            found_last_printed_log = True
                            continue
                        continue
                    if not found_last_printed_log:
                        for log in logs_by_line:
                            print(log)
                last_log_printed = logs_by_line[-1]

            pod = v1.read_namespaced_pod_status(name=pod_name, namespace=namespace)
            if pod.status.phase == PodStatus.FAILED.value:
                print(f"Pod {pod_name} in namespace {namespace} failed.")
                return
            if status:
                if status.value == pod.status.phase:
                    print(f"Pod {pod_name} has reached {status.value} state")
                    return
            time.sleep(1)
    except client.ApiException as e:
        if e.status == 404:
            print(f"Pod {pod_name} not found in namespace {namespace}")
        else:
            raise
    except KeyboardInterrupt:
        print("Log streaming stopped by user")
        return
    except ApiTypeError as e:
        print(f"An error occurred while streaming logs for pod {pod_name} in namespace {namespace}")
        print(e)
        return
    except Exception as e:
        print(f"An unexpected error occurred for pod {pod_name} in namespace {namespace}")
        print(e)
        raise


def get_pod_status_from_job(job_name: str, namespace: str = DEFAULT_NAMESPACE, context_name: str = None) -> Optional[
    str]:
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.BatchV1Api(k8s_api_client)
    try:
        job = v1.read_namespaced_job(name=job_name, namespace=namespace)
        if job.status.failed == 1:
            return PodStatus.FAILED.value
        if job.status.succeeded == 1:
            return PodStatus.SUCCEEDED.value
        return None
    except ApiException as e:
        if e.status == 404:
            print(f"Job {job_name} not found in namespace {namespace}")
            return PodStatus.FAILED.value
        else:
            raise e


def start_streaming_service(service_name, namespace, context_name: Optional[str] = None):
    # Load kube config
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create a Kubernetes API client
    v1 = client.CoreV1Api(k8s_api_client)

    # Stream the service status
    try:
        while True:
            pods = v1.list_namespaced_pod(namespace, label_selector=f'serving.knative.dev/service={service_name}')
            if pods.items:
                sorted_pods = sorted(pods.items, key=lambda x: x.metadata.name, reverse=True)
                print(f"Pods scheduled for service {service_name}: {sorted_pods[0].metadata.name}")
                break
            time.sleep(5)
        # Start streaming the logs from the pod
        thread = Thread(target=start_streaming_pod,
                        args=(sorted_pods[0].metadata.name, namespace, PodStatus.RUNNING, 'user-container'))
        thread.start()
    except client.ApiException as e:
        if e.status == 404:
            print(f"Service {service_name} not found in namespace {namespace}")
        else:
            raise
    except KeyboardInterrupt:
        print("Service status streaming stopped by user")
        return
    except Exception as e:
        print("An unexpected error occurred for service status streaming")
        raise


def check_nodes_ready(label_selector):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    ready_nodes = []
    nodes = v1.list_node(label_selector=label_selector).items
    for node in nodes:
        for condition in node.status.conditions:
            if condition.type == "Ready" and condition.status == "True":
                ready_nodes.append(node.metadata.name)
    return len(ready_nodes) == len(nodes), ready_nodes


def evict_pods_from_node(node_name, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    core_api = client.CoreV1Api(k8s_api_client)
    pods = core_api.list_pod_for_all_namespaces(field_selector=f'spec.nodeName={node_name}').items
    for pod in pods:
        if pod.metadata.owner_references and any(owner.kind == "DaemonSet" for owner in pod.metadata.owner_references):
            continue
        eviction = client.V1Eviction(
            metadata=client.V1ObjectMeta(name=pod.metadata.name, namespace=pod.metadata.namespace))
        retry_attempts = 0
        max_retries = 5
        backoff_delay = 1  # Start with 1 second delay
        while retry_attempts < max_retries:
            try:
                core_api.create_namespaced_pod_eviction(name=pod.metadata.name, namespace=pod.metadata.namespace,
                                                        body=eviction)
                print(f"Evicting pod {pod.metadata.name} from node {node_name}.")
                break  # Eviction successful, break out of the retry loop
            except ApiException as e:
                if e.status == 429:  # Too Many Requests
                    print(
                        f"Rate limit exceeded when evicting pod {pod.metadata.name}: {e}. Retrying in {backoff_delay} seconds...")
                    time.sleep(backoff_delay)
                    backoff_delay *= 2  # Exponential backoff
                    retry_attempts += 1
                else:
                    print(f"Exception when evicting pod {pod.metadata.name}: {e}")
                    break  # Break on any other exception
            except Exception as e:
                print(f"An unexpected error occurred: {e}")
                break
        if retry_attempts == max_retries:
            print(f"Failed to evict pod {pod.metadata.name} after {max_retries} attempts.")


def drain_and_delete_node(node_name, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    core_api = client.CoreV1Api(k8s_api_client)
    # Cordon the node
    body = {"spec": {"unschedulable": True}}
    core_api.patch_node(node_name, body)
    click.echo(f"Cordoned node {node_name}")

    # Evict all pods from the node
    evict_pods_from_node(node_name)
    click.echo(f"All pods evicted from node {node_name}")

    # delete the node
    core_api.delete_node(node_name)


def get_nodes_not_using_bottlerocket(ec2_node_class_label, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)

    # List all nodes
    nodes = v1.list_node().items

    filtered_nodes = []
    for node in nodes:
        labels = node.metadata.labels
        os_image = node.status.node_info.os_image

        # Check if the node belongs to the specified EC2NodeClass
        if labels.get('karpenter.sh/nodepool') == ec2_node_class_label:
            # Check if the node is using Bottlerocket AMI
            if 'bottlerocket' not in os_image.lower():
                filtered_nodes.append(node.metadata.name)

    return filtered_nodes


def delete_pv_using_name(pv_name: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)

    timeout = 60  # Timeout in seconds
    start_time = time.time()

    try:
        v1.delete_persistent_volume(name=pv_name)
        while time.time() - start_time < timeout:
            try:
                v1.read_persistent_volume(name=pv_name)
                print(f"Waiting for PV {pv_name} to be deleted...")
                time.sleep(5)  # Wait for 5 seconds before checking again
            except ApiException as e:
                if e.status == 404:
                    print(f"PersistentVolume {pv_name} successfully deleted.")
                    return
                else:
                    raise
        print(f"Timeout reached while waiting for PV {pv_name} to be deleted.")
    except client.exceptions.ApiException as e:
        if e.status == 404:
            print(f"PersistentVolume {pv_name} not found")
        else:
            raise e


def delete_pvc_using_name_and_namespace(pvc_name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    timeout = 60  # Timeout in seconds
    start_time = time.time()
    try:
        v1.delete_namespaced_persistent_volume_claim(name=pvc_name, namespace=namespace)
        while time.time() - start_time < timeout:
            try:
                v1.read_namespaced_persistent_volume_claim(name=pvc_name, namespace=namespace)
                print(f"Waiting for PVC {pvc_name} to be deleted...")
                time.sleep(5)  # Wait for 5 seconds before checking again
            except ApiException as e:
                if e.status == 404:
                    print(f"PersistentVolumeClaim {pvc_name} successfully deleted from namespace {namespace}.")
                    return
                else:
                    raise
        print(f"Timeout reached while waiting for PVC {pvc_name} to be deleted.")
    except client.exceptions.ApiException as e:
        if e.status == 404:
            print(f"PersistentVolumeClaim {pvc_name} not found in namespace {namespace}")
        else:
            raise e


def check_pvc_exists_by_name(claim_name, namespace, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Create an instance of the CoreV1Api
    v1 = client.CoreV1Api(k8s_api_client)

    try:
        # Attempt to read the specified PVC in the given namespace
        v1.read_namespaced_persistent_volume_claim(name=claim_name, namespace=namespace)
        return True  # PVC exists
    except ApiException as e:
        if e.status == 404:
            return False  # PVC does not exist
        else:
            print(f"An error occurred: {e}")
            raise e


def get_image_tags_to_retain(sanitised_project_name: str, service_name: str, namespace: str,
                             context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)
    click.echo(f"Cleaning up old image tags for project {sanitised_project_name}, service {service_name}.")
    group = "serving.knative.dev"
    version = "v1"
    plural = "revisions"

    # List all revisions for the service
    # Define the label selector string based on the service name
    label_selector = f"serving.knative.dev/service={service_name}"

    # Use the label selector in the list_namespaced_custom_object call to directly filter revisions
    service_revisions = \
        k8s_client.list_namespaced_custom_object(group, version, namespace, plural, label_selector=label_selector)[
            'items']

    # Sort all revisions by configurationGeneration in descending order
    all_revisions_sorted = sorted(service_revisions, key=lambda x: int(
        x['metadata']['labels']['serving.knative.dev/configurationGeneration']), reverse=True)

    # Identify all "Ready" revisions
    ready_revisions = [rev for rev in all_revisions_sorted if any(
        cond['type'] == 'Ready' and cond['status'] == 'True' for cond in rev['status']['conditions'])]

    # Find the index of the latest "Ready" revision
    if ready_revisions:
        latest_ready_revision = ready_revisions[0]
        latest_ready_index = all_revisions_sorted.index(latest_ready_revision)
    else:
        latest_ready_index = -1

    # Determine active revisions
    active_revisions = ready_revisions[:3]  # Last three "Ready" revisions
    if latest_ready_index != -1:
        active_revisions += all_revisions_sorted[:latest_ready_index]  # Any newer revisions

    # Find the index for slicing stale revisions
    if len(ready_revisions) >= 3:
        third_last_ready_index = all_revisions_sorted.index(ready_revisions[2])
    elif len(ready_revisions) == 2:
        third_last_ready_index = all_revisions_sorted.index(ready_revisions[1])
    else:
        third_last_ready_index = latest_ready_index

    # Determine stale revisions, ensuring last three "Ready" revisions are not included
    stale_revisions = all_revisions_sorted[third_last_ready_index + 1:] if third_last_ready_index != -1 else []
    retained_revisions = all_revisions_sorted[
                         :third_last_ready_index + 1] if third_last_ready_index != -1 else all_revisions_sorted

    image_tags_to_retain = []
    image_tags_to_delete = []
    for rev in retained_revisions:
        # check if this exists yaml_dict['spec']['template']['metadata']['annotations']['image_tag'] = image_tag
        if rev['metadata']['annotations'] and 'image_tag' in rev['metadata']['annotations']:
            image_tags_to_retain.append(rev['metadata']['annotations']['image_tag'])
            click.echo(f'Retaining revision {rev["metadata"]["name"]}')
        else:
            print(f'No image tag found for revision {rev["metadata"]["name"]}')

    for rev in stale_revisions:
        # check if this exists yaml_dict['spec']['template']['metadata']['annotations']['image_tag'] = image_tag
        if rev['metadata']['annotations'] and 'image_tag' in rev['metadata']['annotations']:
            image_tags_to_delete.append(rev['metadata']['annotations']['image_tag'])
            click.echo(f'Queuing revision {rev["metadata"]["name"]} for deletion')
        else:
            print(f'No image tag found for revision {rev["metadata"]["name"]}')
        try:
            # Delete the revision
            k8s_client.delete_namespaced_custom_object(group=group, version=version, namespace=namespace, plural=plural,
                                                       name=rev['metadata']['name'], body=client.V1DeleteOptions()
                                                       # You can customize this if needed
                                                       )
        except client.exceptions.ApiException as e:
            print(f"Failed to delete revision {rev['metadata']['name']}: {e}")
        except Exception as e:
            print(f"An unexpected error occurred while deleting revision {rev['metadata']['name']}: {e}")
    return image_tags_to_retain


def get_pod_name_corresponing_to_job(job_name: str, namespace: str = DEFAULT_NAMESPACE,
                                     context_name: Optional[str] = None, max_retries = 5) -> Optional[str]:
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    for attempt in range(max_retries):
        pods = v1.list_namespaced_pod(namespace=namespace)
        for pod in pods.items:
            if not pod.metadata.owner_references:
                continue
            for owner in pod.metadata.owner_references:
                if owner.name == job_name:
                    return pod.metadata.name
        backoff_delay = 2 ** attempt
        time.sleep(backoff_delay)
    return None


def create_new_namespace(env_name: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    namespace = client.V1Namespace()
    namespace.metadata = client.V1ObjectMeta(name=env_name)
    v1 = client.CoreV1Api(k8s_api_client)
    v1.create_namespace(body=namespace)
    click.echo(f"Namespace {env_name} created successfully.")


def delete_namespace(env_name: str, context_name: Optional[str] = None):
    if env_name == DEFAULT_NAMESPACE or env_name in TENSORFUSE_NAMESPACES:
        click.echo(f"Cannot delete namespace {env_name}")
        return
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    v1.delete_namespace(name=env_name)
    click.echo(f"Namespace {env_name} deleted successfully.")


def list_all_namespaces(context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    namespaces = v1.list_namespace().items
    return [namespace.metadata.name for namespace in namespaces]


def ssh_into_pod(pod_name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    
    command = ["kubectl", "--context", f"{context_name}", "exec", "-it", pod_name, "-n", namespace, "-c",
               "user-container", "--", "sh"]
    subprocess.run(command)

def ssh_into_pod_with_podman(pod_name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    command = ["kubectl", "--context", f"{context_name}", "exec", "-it", pod_name, "-n", namespace, "-c",
               "user-container", "--", "sh", "-c", "podman exec -it $(podman ps -q) /bin/bash"]
    subprocess.run(command)


def get_tensorkube_cluster_context_name() -> Optional[str]:
    # Get the list of all contexts and the current context
    contexts, current_context = config.list_kube_config_contexts()

    tensorkube_contexts = [context['name'] for context in contexts if (
            f"cluster/{get_cluster_name()}" in context['name'] or f"@{get_cluster_name()}" in context['name'])]
    if len(tensorkube_contexts) == 1:
        return tensorkube_contexts[0]
    elif len(tensorkube_contexts) > 1:
        valid_context = remove_invalid_contexts(tensorkube_contexts)
        return valid_context
    else:
        click.echo(click.style(f"No context found for {get_cluster_name()} cluster.", fg="red"))
        click.echo(
            f"If you have already created a {get_cluster_name()} cluster and have access to it, please run the command")
        click.echo(click.style("tensorkube sync", fg="cyan"))
        click.echo("Otherwise, please create a new cluster using the command")
        click.echo(click.style("tensorkube configure", fg="cyan"))
        return None


def remove_invalid_contexts(tensorkube_contexts: List[str]) -> Optional[str]:
    for context in tensorkube_contexts:
        try:
            command = ["kubectl", "--context", f"{context}", "get", "pods"]
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError:
            command = ["kubectl", "config", "delete-context", context]
            subprocess.run(command)
            tensorkube_contexts.remove(context)

    if len(tensorkube_contexts) == 1:
        return tensorkube_contexts[0]
    elif len(tensorkube_contexts) > 1:
        return tensorkube_contexts[0]
    return None


def create_secret(name: str, namespace: str, data: Dict[str, str], force: bool = False, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
 
    secret = client.V1Secret()
    secret.metadata = client.V1ObjectMeta(name=name, namespace=namespace)
    secret.string_data = data


    try:
        # Check if the secret already exists
        v1.read_namespaced_secret(name, namespace)
        if force:
            # If the secret exists and force is True, update the secret
            v1.replace_namespaced_secret(name, namespace, body=secret)
            click.echo(click.style("Updated already existing secret ", fg=CliColors.SUCCESS.value) 
                       + click.style(name, bold=True, fg=CliColors.SUCCESS.value)
                       + click.style(" in namespace ", fg=CliColors.SUCCESS.value) 
                       + click.style(namespace, bold=True, fg=CliColors.SUCCESS.value))
        else:
            click.echo(click.style("Secret ", fg=CliColors.WARNING.value) 
                       + click.style(name,  bold=True, fg=CliColors.WARNING.value)
                       + click.style(" already exists in namespace ", fg=CliColors.WARNING.value)
                       + click.style(namespace, bold=True, fg=CliColors.WARNING.value)
                       + click.style(". Use --force to update the secret.", fg=CliColors.WARNING.value))
    except ApiException as e:
        if e.status == 404:
            # If the secret does not exist, create it
            try:
                v1.create_namespaced_secret(namespace=namespace, body=secret)
                click.echo(click.style("Secret ", fg=CliColors.SUCCESS.value) 
               + click.style(name, bold=True, fg=CliColors.SUCCESS.value)
               + click.style(" created successfully.", fg=CliColors.SUCCESS.value))
            except ApiException as e:
                click.echo(click.style("An error occurred while creating the secret", fg=CliColors.ERROR.value))
                click.echo(click.style(f"Error: {e}", fg=CliColors.ERROR.value))
        else:
            raise e

def list_secrets(namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return []
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    secrets = v1.list_namespaced_secret(namespace).items
    if not secrets:
        return []
    return secrets


def delete_secret(name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    try:
        v1.delete_namespaced_secret(name, namespace)
        click.echo(click.style("Secret ", fg=CliColors.SUCCESS.value) 
                   + click.style(name, bold=True, fg=CliColors.SUCCESS.value)
                   + click.style(" deleted successfully.", fg=CliColors.SUCCESS.value))
    except ApiException as e:
        if e.status == 404:
            click.echo(click.style("Secret ", fg=CliColors.WARNING.value) 
                       + click.style(name, bold=True, fg=CliColors.WARNING.value)
                       + click.style(" not found in namespace ", fg=CliColors.WARNING.value) 
                       + click.style(namespace, bold=True, fg=CliColors.WARNING.value))
        else:
            raise e


def create_configmap(name: str, namespace: str, data: Dict[str, str], force: bool = False, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client) 
    configmap = client.V1ConfigMap()
    configmap.metadata = client.V1ObjectMeta(name=name, namespace=namespace)
    configmap.data = {"config": yaml.dump(data)}
    try:
        # Check if the configmap already exists
        v1.read_namespaced_config_map(name, namespace)
        if force:
            # If the configmap exists and force is True, update the configmap
            v1.replace_namespaced_config_map(name, namespace, body=configmap)
    except ApiException as e:
        if e.status == 404:
            # If the configmap does not exist, create it
            try:
                v1.create_namespaced_config_map(namespace=namespace, body=configmap)
            except ApiException as e:
                raise e
        else:
            raise e

def get_pods_for_jobs(job_name: str, namespace: str = "default"):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    api = client.CoreV1Api(k8s_api_client)
    try:
        return api.list_namespaced_pod(
            namespace=namespace,
            label_selector=f"job-name={job_name}"
        )
    except ApiException as e:
        print(f"Exception when calling CoreV1Api->list_namespaced_pod: {e}")
        return None


def list_jobs(namespace: Optional[str] = None, all: bool = False,
              job_name_prefix: Optional[str] = None):
    """
    Lists Kubernetes jobs in a namespace or across all namespaces with optional job name prefix filtering

    Args:
        namespace: Specific namespace to list jobs from
        all: If True, lists jobs from all namespaces
        context_name: Kubernetes context name to use
        job_name_prefix: Optional prefix to filter job names

    Returns:
        Dict containing list of jobs or None if context not found
    """
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None

    api_client = config.new_client_from_config(context=context_name)
    batch_v1 = client.BatchV1Api(api_client)

    try:
        if all:
            job_list = batch_v1.list_job_for_all_namespaces()
        else:
            namespace = namespace if namespace else DEFAULT_NAMESPACE
            job_list = batch_v1.list_namespaced_job(namespace=namespace)

        # Filter jobs by prefix if specified
        if job_name_prefix:
            filtered_items = [
                job for job in job_list.items
                if job.metadata.name.startswith(job_name_prefix)
            ]
            job_list.items = filtered_items

        return job_list
    except ApiException as e:
        print(f"Exception when calling BatchV1Api: {e}")
        return None


def delete_job(job_name: str, namespace: str = "default") -> bool:
    """
    Stops and deletes a Kubernetes job and its associated pods

    Args:
        job_name: Name of the job to delete
        namespace: Kubernetes namespace where the job exists

    Returns:
        bool: True if deletion was successful, False otherwise
    """
    try:
        # Get API clients
        api_client = config.new_client_from_config(context=get_tensorkube_cluster_context_name())
        batch_v1 = client.BatchV1Api(api_client)
        core_v1 = client.CoreV1Api(api_client)

        # First check if job exists
        try:
            batch_v1.read_namespaced_job(name=job_name, namespace=namespace)
        except ApiException as e:
            if e.status == 404:
                click.echo(click.style(f"Job {job_name} not found in namespace {namespace}", fg='red'))
                return False
            raise e

        # Delete associated pods first
        try:
            pods = core_v1.list_namespaced_pod(
                namespace=namespace,
                label_selector=f"job-name={job_name}"
            )
            for pod in pods.items:
                core_v1.delete_namespaced_pod(
                    name=pod.metadata.name,
                    namespace=namespace
                )
        except ApiException as e:
            print(f"Error deleting pods for job {job_name}: {e}")

        # Delete the job
        try:
            batch_v1.delete_namespaced_job(
                name=job_name,
                namespace=namespace,
                body=client.V1DeleteOptions(
                    propagation_policy='Foreground',
                    grace_period_seconds=0
                )
            )
            click.echo(click.style(f"Successfully deleted job {job_name}", fg='green'))
            return True
        except ApiException as e:
            click.echo(click.style(f"Error deleting job {job_name}: {e}", fg='red'))
            return False

    except Exception as e:
        print(f"Error in delete_job: {e}")
        click.echo(click.style(f"Failed to delete job: {str(e)}", fg='red'))
        return False


def list_keda_scaled_jobs():
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None

    api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(api_client)

    try:
        keda_scaled_jobs = k8s_client.list_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            plural="scaledjobs",
            namespace="keda"
        )
        return keda_scaled_jobs
    except ApiException as e:
        print(f"Exception when calling CustomObjectsApi: {e}")
        return None


def list_trigger_authentications():
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None

    api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(api_client)

    try:
        trigger_authentications = k8s_client.list_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            plural="triggerauthentications",
            namespace="keda"
        )
        return trigger_authentications
    except ApiException as e:
        print(f"Exception when calling CustomObjectsApi: {e}")
        return None


def delete_trigger_authentication(trigger_auth_name: str):
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        return None

    api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(api_client)

    try:
        k8s_client.delete_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            plural="triggerauthentications",
            namespace="keda",
            name=trigger_auth_name
        )
    except ApiException as e:
        print(f"Exception when calling CustomObjectsApi: {e}")
        return None

def add_user_to_aws_auth_group(user_name, group_name):
    v1 = client.CoreV1Api()
    # details of the aws-auth ConfigMap
    namespace = "kube-system"
    configmap_name = "aws-auth"

    configmap = v1.read_namespaced_config_map(configmap_name, namespace)
    map_users = yaml.safe_load(configmap.data.get("mapUsers", "[]"))
    # Append the new user to the mapUsers list
    new_user = {
        "userarn": f"arn:aws:iam::{get_aws_account_id()}:user/{user_name}",
        "username": user_name,
        "groups": [group_name]
    }

    map_users.append(new_user)
    # Update the mapUsers back into the ConfigMap data
    configmap.data["mapUsers"] = yaml.dump(map_users)

    # Update the ConfigMap with the new user
    v1.replace_namespaced_config_map(name=configmap_name, namespace=namespace, body=configmap)


def add_user_to_k8s(user_name):
    config.load_kube_config()
    rbac_api = client.RbacAuthorizationV1Api()
    cluster_role_name = f"{get_cluster_name()}-{user_name}-job-cluster-role"
    cluster_role = client.V1ClusterRole(
        metadata=client.V1ObjectMeta(name=cluster_role_name),
        rules=[
            client.V1PolicyRule(
                api_groups=["keda.sh"],
                resources=["scaledjobs"],
                verbs=["create", "get", "list", "delete"]
            )
        ]
    )
    try:
        rbac_api.create_cluster_role(body=cluster_role)
    except ApiException as e:
        if e.status == 409:
            print(f"ClusterRole {cluster_role_name} already exists.")
        else:
            raise
    print(f"ClusterRole {cluster_role_name} created successfully.")
    group_name = f"{get_cluster_name()}-{user_name}-job-access"
    cluster_role_binding_name = f"{get_cluster_name()}-{user_name}-job-cluster-role-binding"
    # Define ClusterRoleBinding to bind the role to a user
    cluster_role_binding = client.V1ClusterRoleBinding(
        metadata=client.V1ObjectMeta(name=cluster_role_binding_name),
        role_ref=client.V1RoleRef(
            api_group="rbac.authorization.k8s.io",
            kind="ClusterRole",
            name=cluster_role_name
        ),
        subjects=[
            client.RbacV1Subject(
                kind="Group",
                name=group_name,
            )
        ]
    )
    try:
        rbac_api.create_cluster_role_binding(body=cluster_role_binding)
    except ApiException as e:
        if e.status == 409:
            print(f"ClusterRoleBinding {cluster_role_binding_name} already exists.")
        else:
            raise
    print(f"ClusterRoleBinding {cluster_role_binding_name} created successfully.")
    add_user_to_aws_auth_group(user_name, group_name)
    return



def generate_kubeconfig():
    eks_client = boto3.client(
        'eks',
        region_name=REGION,
    )

    cluster_name = get_cluster_name()
    # Retrieve cluster information
    cluster_info = eks_client.describe_cluster(name=get_cluster_name())['cluster']

    # Extract cluster details
    endpoint = cluster_info['endpoint']
    cluster_ca_certificate = cluster_info['certificateAuthority']['data']


    # Create a kubeconfig dictionary
    kubeconfig_data = {
        "apiVersion": "v1",
        "clusters": [{
            "cluster": {
                "server": endpoint,
                "certificate-authority-data": cluster_ca_certificate
            },
            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
        }],
        "contexts": [{
            "context": {
                "cluster": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
                "user": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
            },
            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
        }],
        "current-context": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
        "kind": "Config",
        "users": [{
            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
            "user": {
                "exec": {
                    "apiVersion": "client.authentication.k8s.io/v1beta1",
                    "command": "aws",
                    "args": [
                        "eks",
                        "get-token",
                        "--cluster-name", cluster_name
                    ],
                    "env": [
                        {"name": "AWS_DEFAULT_REGION", "value": REGION}
                    ]
                }
            }
        }]
    }
    kubeconfig_dir = os.path.expanduser('~/.kube')
    kubeconfig_file = os.path.join(kubeconfig_dir, 'config')

    # Create the directory if it does not exist
    if not os.path.exists(kubeconfig_dir):
        os.makedirs(kubeconfig_dir,exist_ok=True)

    # Load existing kubeconfig or create a new one if it doesn't exist
    if os.path.exists(kubeconfig_file):
        with open(kubeconfig_file, 'r') as f:
            existing_config = yaml.safe_load(f) or {}
    else:
        existing_config = {'apiVersion': 'v1', 'kind': 'Config', 'clusters': [], 'contexts': [], 'users': []}

    # Merge new kubeconfig entry into existing config
    # Function to update or add an entry in the list
    def update_or_add_entry(existing_list, new_entry, key):
        for i, entry in enumerate(existing_list):
            if entry[key] == new_entry[key]:
                existing_list[i] = new_entry  # Update the existing entry
                return
        existing_list.append(new_entry)  # Add as a new entry if not found

    # Merge new kubeconfig entry into existing config
    update_or_add_entry(existing_config['clusters'], kubeconfig_data['clusters'][0], 'name')
    update_or_add_entry(existing_config['contexts'], kubeconfig_data['contexts'][0], 'name')
    update_or_add_entry(existing_config['users'], kubeconfig_data['users'][0], 'name')
    existing_config['current-context'] = kubeconfig_data['current-context']


    # Write back to the default kubeconfig file
    with open(kubeconfig_file, 'w') as f:
        yaml.dump(existing_config, f)

def create_service_account(name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")

    k8s_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_client)
    service_account = client.V1ServiceAccount()
    service_account.metadata = client.V1ObjectMeta(name=name)
    try:
        v1.create_namespaced_service_account(namespace=namespace, body=service_account)
    except ApiException as e:
        raise e
    

def create_service_account_with_role_arn(name: str, namespace: str, role_arn: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")

    k8s_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_client)

    # Check if the ServiceAccount already exists
    try:
        v1.read_namespaced_service_account(name=name, namespace=namespace)
        print(f"ServiceAccount '{name}' already exists in namespace '{namespace}'.")
    except ApiException as e:
        if e.status == 404:
            # ServiceAccount does not exist; create it
            service_account = client.V1ServiceAccount(
                metadata=client.V1ObjectMeta(
                    name=name,
                    annotations={"eks.amazonaws.com/role-arn": role_arn},
                    namespace=namespace
                ),
            )
            try:
                v1.create_namespaced_service_account(namespace=namespace, body=service_account)
                print(f"ServiceAccount '{name}' created in namespace '{namespace}'.")
            except ApiException as create_error:
                print(f"Failed to create ServiceAccount '{name}': {create_error}")
                raise create_error
        else:
            print(f"Failed to check ServiceAccount existence: {e}")
            raise e



def create_k8s_role(role_name: str, namespace: str, rules: List[Dict[str, Any]], context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")

    k8s_client = config.new_client_from_config(context=context_name)
    rbac_api = client.RbacAuthorizationV1Api(k8s_client)
    role = client.V1Role()
    role.metadata = client.V1ObjectMeta(name=role_name)
    role.rules = rules
    try:
        rbac_api.create_namespaced_role(namespace=namespace, body=role)
    except ApiException as e:
        raise e

def create_role_binding(rb_name: str, rb_namespace: str, role_name: str, sa_name: str, sa_namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")

    k8s_api_client = config.new_client_from_config(context=context_name)
    rbac_api = client.RbacAuthorizationV1Api(k8s_api_client)
    role_binding = client.V1RoleBinding(
        role_ref=client.V1RoleRef(api_group="rbac.authorization.k8s.io", kind="Role", name=role_name),
        subjects=[client.RbacV1Subject(kind="ServiceAccount", name=sa_name, namespace=sa_namespace)])
    role_binding.metadata = client.V1ObjectMeta(name=rb_name)
    try:
        rbac_api.create_namespaced_role_binding(namespace=rb_namespace, body=role_binding)
    except ApiException as e:
        raise e


def is_existing_scaled_job(job_name: str, namespace: str, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)
    try:
        job = k8s_client.get_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            plural="scaledjobs",
            name=job_name,
            namespace=namespace
        )
        return True, job
    except ApiException as e:
        if e.status == 404:
            return False, None
        raise e


def apply_domain_claim(domain: str, namespace: str):
    k8s_api_client = config.new_client_from_config(context=get_tensorkube_cluster_context_name())
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    config_yaml_path = resource_filename('tensorkube', 'configurations/build_configs/domain_claim_base_config.yaml')

    with open(config_yaml_path, 'r') as f:
        config_yaml = yaml.safe_load(f)

    config_yaml['metadata']['name'] = domain
    config_yaml['spec']['namespace'] = namespace

    group = 'networking.internal.knative.dev'
    version = 'v1alpha1'
    plural = 'clusterdomainclaims'

    try:
        # Use cluster-scoped methods instead of namespaced ones
        existing_claim = k8s_client.get_cluster_custom_object(
            group=group, version=version,
            plural=plural, name=domain
        )
        resource_version = existing_claim['metadata']['resourceVersion']
        config_yaml['metadata']['resourceVersion'] = resource_version
        k8s_client.patch_cluster_custom_object(
            group=group, version=version,
            plural=plural, name=domain,
            body=config_yaml
        )
        click.echo(f"Updated ClusterDomainClaim for domain {domain}")
    except ApiException as e:
        if e.status == 404:
            k8s_client.create_cluster_custom_object(
                group=group, version=version,
                plural=plural, body=config_yaml
            )
            click.echo(f"Created ClusterDomainClaim for domain {domain}")
        else:
            raise e
        


def patch_service_account(role_arn, service_account_name, namespace):
    # Load the Kubernetes configuration
    config.load_kube_config()

    # Create the Kubernetes API client
    v1 = client.CoreV1Api()

    # Define the patch for the ServiceAccount
    patch_body = {
        "metadata": {
            "annotations": {
                "eks.amazonaws.com/role-arn": role_arn
            }
        }
    }

    try:
        # Patch the existing ServiceAccount
        response = v1.patch_namespaced_service_account(
            name=service_account_name,
            namespace=namespace,
            body=patch_body
        )
        print(f"ServiceAccount '{service_account_name}' in namespace '{namespace}' patched successfully with role ARN.")
    except client.exceptions.ApiException as e:
        print(f"Failed to patch ServiceAccount: {e.reason}")


def get_jobs_by_container_image(image: str, namespace: str = "keda") -> list:
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise Exception("Unable to get cluster context")
    k8s_api_client = config.new_client_from_config(context=context_name)
    k8s_client = client.CustomObjectsApi(k8s_api_client)

    try:
        scaled_jobs = k8s_client.list_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            plural="scaledjobs",
            namespace=namespace
        )
        matching_jobs = []
        for job in scaled_jobs.get("items", []):
            containers = job.get("spec", {}).get("jobTargetRef", {}).get("template", {}).get("spec", {}).get(
                "containers", [])
            if any(container.get("image") == image for container in containers):
                matching_jobs.append(job)
        return matching_jobs

    except ApiException as e:
        if e.status == 404:
            return []
        raise e
